package org.artifactory.repo.migration.mappers;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RemoteRepositoryRepoTypeConfig;
import org.artifactory.repo.CdnRedirectRepoConfig;
import org.artifactory.repo.ChecksumPolicyType;
import org.artifactory.repo.DownloadRedirectRepoConfig;
import org.artifactory.repo.remote.ContentSynchronisation;
import org.artifactory.repo.remote.PropertiesContent;
import org.artifactory.repo.remote.SourceContent;
import org.artifactory.repo.remote.StatisticsContent;
import org.artifactory.repo.vcs.VcsGitProvider;
import org.artifactory.repo.vcs.VcsType;

import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;

@Slf4j
public class RemoteRepoConfigMapper {

    public static RemoteRepoConfig mapRemoteRepoConfig(RemoteRepoConfig remoteRepoConfig,
            Map<String, Object> newValues, CentralConfigDescriptor descriptor) {
        RepoConfigMigrationPatchUtils.buildBaseConfig(remoteRepoConfig.getBaseConfig(), newValues, descriptor);
        buildRepoTypeConfig(remoteRepoConfig.getRepoTypeConfig(), newValues, descriptor);
        buildPackageTypeConfig(remoteRepoConfig.getPackageTypeConfig(), remoteRepoConfig.getPackageType(),
                newValues);
        RepoConfigMigrationPatchUtils.buildSecurityConfig(remoteRepoConfig.getSecurityConfig(), newValues, descriptor);
        return remoteRepoConfig;
    }

    private static Map<String, String> buildPackageTypeConfig(Map<String, String> packageTypeConfig,
            RepoType repoType, Map<String, Object> newValues) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if (repoType.isMavenGroup()) {
                    addValueToPackageTypeConfigIfNeeded("suppressPomConsistencyChecks", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("fetchJarsEagerly", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("fetchSourcesEagerly", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("rejectInvalidJars", packageTypeConfig, newValue);
                    if (RepoType.P2.equals(repoType)) {
                        addValueToPackageTypeConfigIfNeeded("p2OriginalUrl", packageTypeConfig, newValue);
                        addValueToPackageTypeConfigIfNeeded("forceP2Authentication", packageTypeConfig, newValue);
                    }
                } else if (RepoType.Debian.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("ddebSupported", packageTypeConfig, newValue);
                } else if (RepoType.Cargo.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("cargoAnonymousAccess", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("cargoInternalIndex", packageTypeConfig, newValue);
                    if ("git".equals(newValue.getKey())) {
                        Map<String, Object> gitNewValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> gitNewValue : gitNewValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("gitRegistryUrl", packageTypeConfig,
                                    gitNewValue);
                        }
                    }
                } else if (RepoType.Bower.equals(repoType)) {
                    if ("bower".equals(newValue.getKey())) {
                        packageTypeConfig.put("bowerRegistryUrl",
                                String.valueOf(((Map<?, ?>) newValue.getValue()).get("bowerRegistryUrl")));
                    }
                    if ("vcs".equals(newValue.getKey())) {
                        setVcsNeeded((Map<String, Object>) newValue.getValue(), packageTypeConfig);
                    }
                } else if (RepoType.Composer.equals(repoType)) {
                    if ("composer".equals(newValue.getKey())) {
                        Map<String, Object> composerNewValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> composerNewValue : composerNewValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("composerRegistryUrl", packageTypeConfig,
                                    composerNewValue);
                        }
                    }
                    if ("vcs".equals(newValue.getKey())) {
                        setVcsNeeded((Map<String, Object>) newValue.getValue(), packageTypeConfig);
                    }
                } else if (RepoType.Docker.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("dockerTagRetention", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("dockerApiVersion", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("blockPushingSchema1", packageTypeConfig, newValue);
                    addValueToPackageTypeConfigIfNeeded("maxUniqueTags", packageTypeConfig, newValue);
                    RepoConfigMigrationPatchUtils.externalDependenciesToConfig(packageTypeConfig, newValues);
                } else if (RepoType.Helm.equals(repoType)) {
                    RepoConfigMigrationPatchUtils.externalDependenciesToConfig(packageTypeConfig, newValues);
                } else if (RepoType.NuGet.equals(repoType)) {
                    if ("nuget".equals(newValue.getKey())) {
                        Map<String, Object> nugetValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> nugetValue : nugetValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("downloadContextPath", packageTypeConfig, nugetValue);
                            addValueToPackageTypeConfigIfNeeded("v3FeedUrl", packageTypeConfig, nugetValue);
                            addValueToPackageTypeConfigIfNeeded("feedContextPath", packageTypeConfig, nugetValue);
                        }
                    }
                    addValueToPackageTypeConfigIfNeeded("forceNugetAuthentication", packageTypeConfig, newValue);
                } else if (RepoType.Pypi.equals(repoType)) {
                    if ("pypi".equals(newValue.getKey())) {
                        Map<String, Object> pypiValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> pypiValue : pypiValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("pyPIRegistryUrl", packageTypeConfig, pypiValue);
                            addValueToPackageTypeConfigIfNeeded("packagesContextPath", packageTypeConfig, pypiValue);
                            addValueToPackageTypeConfigIfNeeded("repositorySuffix", packageTypeConfig, pypiValue);
                            addValueToPackageTypeConfigIfNeeded("indexContextPath", packageTypeConfig, pypiValue);
                        }
                    }
                } else if (RepoType.VCS.equals(repoType)) {
                    if ("vcs".equals(newValue.getKey())) {
                        setVcsNeeded((Map<String, Object>) newValue.getValue(), packageTypeConfig);
                    }
                } else if (RepoType.Terraform.equals(repoType)) {
                    if ("terraform".equals(newValue.getKey())) {
                        Map<String, Object> terraformValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> terraformValue : terraformValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("terraformProvidersUrl", packageTypeConfig,
                                    terraformValue);
                            addValueToPackageTypeConfigIfNeeded("terraformRegistryUrl", packageTypeConfig,
                                    terraformValue);
                        }
                    }
                } else if (RepoType.Go.equals(repoType)) {
                    if ("vcs".equals(newValue.getKey())) {
                        setVcsNeeded((Map<String, Object>) newValue.getValue(), packageTypeConfig);
                    }
                    RepoConfigMigrationPatchUtils.externalDependenciesToConfig(packageTypeConfig, newValues);
                } else if (RepoType.Conan.equals(repoType)) {
                    addValueToPackageTypeConfigIfNeeded("forceConanAuthentication", packageTypeConfig, newValue);
                } else if (RepoType.Swift.equals(repoType)) {
                    RepoConfigMigrationPatchUtils.externalDependenciesToConfig(packageTypeConfig, newValues);
                } else if (RepoType.CocoaPods.equals(repoType)) {
                    if ("cocoaPodsConfiguration".equals(newValue.getKey())) {
                        Map<String, Object> cocoaPodsConfigurationValues = (Map<String, Object>) newValue.getValue();
                        for (Map.Entry<String, Object> cocoaPodsConfigurationValue :
                                cocoaPodsConfigurationValues.entrySet()) {
                            addValueToPackageTypeConfigIfNeeded("cocoaPodsSpecsRepoUrl", packageTypeConfig,
                                    cocoaPodsConfigurationValue);
                            if ("specRepoProvider".equals(cocoaPodsConfigurationValue.getKey())) {
                                Map<String, Object> specRepoProviderValues = (Map<String, Object>) newValue.getValue();
                                for (Map.Entry<String, Object> specRepoProviderValue :
                                        specRepoProviderValues.entrySet()) {
                                    addValueToPackageTypeConfigIfNeeded("specsRepoDownloadUrl", packageTypeConfig,
                                            specRepoProviderValue);
                                    addValueToPackageTypeConfigIfNeeded("specsRepoDownloadProvider", packageTypeConfig,
                                            specRepoProviderValue);
                                }
                            }
                        }
                    }
                }
                if ("vcs".equals(newValue.getKey())) {
                    setVcsNeeded((Map<String, Object>) newValue.getValue(), packageTypeConfig);
                }
                RepoConfigMigrationPatchUtils.externalDependenciesToConfig(packageTypeConfig, newValues);
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
        return packageTypeConfig;
    }

    private static void setVcsNeeded(Map<String, Object> vcs, Map<String, String> packageTypeConfig) {
        if (vcs.get("type") != null) {
            packageTypeConfig.put("vcsType",
                    Objects.requireNonNull(VcsType.fromString(String.valueOf(vcs.get("type"))))
                            .name());
        }
        if (vcs.get("git") != null) {
            Map<String, Object> git = (Map<String, Object>) vcs.get("git");
            if (git.get("provider") != null) {
                packageTypeConfig.put("vcsGitProvider",
                        Objects.requireNonNull(VcsGitProvider.fromString(String.valueOf(git.get("provider"))))
                                .name());
            }
            if (git.get("downloadUrl") != null) {
                packageTypeConfig.put("vcsGitDownloadUrl", String.valueOf(git.get("downloadUrl")));
            }
        }
    }

    private static void buildRepoTypeConfig(RemoteRepositoryRepoTypeConfig repoTypeConfig,
            Map<String, Object> newValues, CentralConfigDescriptor descriptor) {
        for (Map.Entry<String, Object> newValue : newValues.entrySet()) {
            try {
                if ("handleReleases".equals(newValue.getKey())) {
                    repoTypeConfig.setHandleReleases((Boolean) newValue.getValue());
                } else if ("maxUniqueSnapshots".equals(newValue.getKey())) {
                    repoTypeConfig.setMaxUniqueSnapshots((int) newValue.getValue());
                } else if ("handleSnapshots".equals(newValue.getKey())) {
                    repoTypeConfig.setHandleSnapshots((Boolean) newValue.getValue());
                } else if ("cdnRedirectRepoConfig".equals(newValue.getKey())) {
                    CdnRedirectRepoConfig cdnRedirectRepoConfig = new CdnRedirectRepoConfig();
                    cdnRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repoTypeConfig.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
                } else if ("downloadRedirect".equals(newValue.getKey())) {
                    DownloadRedirectRepoConfig downloadRedirectRepoConfig = new DownloadRedirectRepoConfig();
                    downloadRedirectRepoConfig.setEnabled(
                            (boolean) ((Map<String, Object>) newValue.getValue()).get("enabled"));
                    repoTypeConfig.setDownloadRedirectConfig(downloadRedirectRepoConfig);
                } else if ("username".equals(newValue.getKey())) {
                    repoTypeConfig.setUsername(String.valueOf(newValue.getValue()));
                } else if ("password".equals(newValue.getKey())) {
                    repoTypeConfig.setPassword(String.valueOf(newValue.getValue()));
                } else if ("proxy".equals(newValue.getKey())) {
                    String proxyRef = String.valueOf(newValue.getValue());
                    if (descriptor.getProxy(proxyRef) == null) {
                        log.debug("Proxy '{}' does not exist in the config descriptor", proxyRef);
                    } else {
                        repoTypeConfig.setProxyRef(proxyRef);
                    }
                } else if ("disableProxy".equals(newValue.getKey())) {
                    repoTypeConfig.setDisableProxy((Boolean) newValue.getValue());
                } else if ("propertySets".equals(newValue.getKey())) {
                    Set<String> existingPropertySetKeys = descriptor.getPropertySets().stream()
                            .map(PropertySet::getName)
                            .collect(Collectors.toSet());
                    Map<String, Object> properties = (Map<String, Object>) newValue.getValue();
                    List<String> propertiesList = properties.entrySet().stream()
                            .sorted(Map.Entry.comparingByKey())
                            .map(Map.Entry::getValue)
                            .map(String.class::cast)
                            .filter(propertySetKey -> {
                                if (!existingPropertySetKeys.contains(propertySetKey)) {
                                    log.debug("Property set '{}' does not exist in the config descriptor",
                                            propertySetKey);
                                    return false;
                                }
                                return true;
                            })
                            .toList();
                    repoTypeConfig.setPropertySetRefs(propertiesList);
                } else if ("remoteRepoChecksumPolicyType".equals(newValue.getKey())) {
                    repoTypeConfig.setChecksumPolicyType(ChecksumPolicyType.fromString(
                            String.valueOf(newValue.getValue())));
                } else if ("assumedOfflinePeriodSecs".equals(newValue.getKey())) {
                    repoTypeConfig.setAssumedOfflinePeriodSecs(((Integer) newValue.getValue()).longValue());
                } else if ("retrievalCachePeriodSecs".equals(newValue.getKey())) {
                    repoTypeConfig.setRetrievalCachePeriodSecs(((Integer) newValue.getValue()).longValue());
                } else if ("missedRetrievalCachePeriodSecs".equals(newValue.getKey())) {
                    repoTypeConfig.setMissedRetrievalCachePeriodSecs(((Integer) newValue.getValue()).longValue());
                } else if ("metadataRetrievalTimeoutSecs".equals(newValue.getKey())) {
                    repoTypeConfig.setMetadataRetrievalTimeoutSecs(((Integer) newValue.getValue()).longValue());
                } else if ("socketTimeoutMillis".equals(newValue.getKey())) {
                    repoTypeConfig.setSocketTimeoutMillis((Integer) newValue.getValue());
                } else if ("unusedArtifactsCleanupPeriodHours".equals(newValue.getKey())) {
                    repoTypeConfig.setUnusedArtifactsCleanupPeriodHours((Integer) newValue.getValue());
                } else if ("clientTlsCertificate".equals(newValue.getKey())) {
                    repoTypeConfig.setClientTlsCertificate(String.valueOf(newValue.getValue()));
                } else if ("queryParams".equals(newValue.getKey())) {
                    repoTypeConfig.setQueryParams(String.valueOf(newValue.getValue()));
                } else if ("url".equals(newValue.getKey())) {
                    repoTypeConfig.setUrl(String.valueOf(newValue.getValue()));
                } else if ("localAddress".equals(newValue.getKey())) {
                    repoTypeConfig.setLocalAddress(String.valueOf(newValue.getValue()));
                } else if ("mismatchingMimeTypesOverrideList".equals(newValue.getKey())) {
                    repoTypeConfig.setMismatchingMimeTypesOverrideList(String.valueOf(newValue.getValue()));
                } else if ("synchronizeProperties".equals(newValue.getKey())) {
                    repoTypeConfig.setSynchronizeProperties((Boolean) newValue.getValue());
                } else if ("enableCookieManagement".equals(newValue.getKey())) {
                    repoTypeConfig.setEnableCookieManagement((Boolean) newValue.getValue());
                } else if ("blackedOut".equals(newValue.getKey())) {
                    repoTypeConfig.setBlackedOut((Boolean) newValue.getValue());
                } else if ("allowAnyHostAuth".equals(newValue.getKey())) {
                    repoTypeConfig.setAllowAnyHostAuth((Boolean) newValue.getValue());
                } else if ("blockMismatchingMimeTypes".equals(newValue.getKey())) {
                    repoTypeConfig.setBlockMismatchingMimeTypes((Boolean) newValue.getValue());
                } else if ("bypassHeadRequests".equals(newValue.getKey())) {
                    repoTypeConfig.setBypassHeadRequests((Boolean) newValue.getValue());
                } else if ("disableUrlNormalization".equals(newValue.getKey())) {
                    repoTypeConfig.setDisableUrlNormalization((Boolean) newValue.getValue());
                } else if ("listRemoteFolderItems".equals(newValue.getKey())) {
                    repoTypeConfig.setListRemoteFolderItems((Boolean) newValue.getValue());
                } else if ("offline".equals(newValue.getKey())) {
                    repoTypeConfig.setOffline((Boolean) newValue.getValue());
                } else if ("enableTokenAuthentication".equals(newValue.getKey())) {
                    repoTypeConfig.setEnableTokenAuthentication((Boolean) newValue.getValue());
                } else if ("priorityResolution".equals(newValue.getKey())) {
                    repoTypeConfig.setPriorityResolution((Boolean) newValue.getValue());
                } else if ("shareConfiguration".equals(newValue.getKey())) {
                    repoTypeConfig.setShareConfiguration((Boolean) newValue.getValue());
                } else if ("storeArtifactsLocally".equals(newValue.getKey())) {
                    repoTypeConfig.setStoreArtifactsLocally((Boolean) newValue.getValue());
                } else if ("hardFail".equals(newValue.getKey())) {
                    repoTypeConfig.setHardFail((Boolean) newValue.getValue());
                } else if ("archiveBrowsingEnabled".equals(newValue.getKey())) {
                    repoTypeConfig.setArchiveBrowsingEnabled((Boolean) newValue.getValue());
                } else if ("curated".equals(newValue.getKey())) {
                    repoTypeConfig.setCurated((Boolean) newValue.getValue());
                } else if ("propagateQueryParams".equals(newValue.getKey())) {
                    repoTypeConfig.setPropagateQueryParams((Boolean) newValue.getValue());
                } else if ("contentSynchronisation".equals(newValue.getKey())) {
                    ContentSynchronisation contentSynchronisation = new ContentSynchronisation();
                    if (repoTypeConfig.getContentSynchronisation() != null) {
                        contentSynchronisation = repoTypeConfig.getContentSynchronisation();
                    }
                    Map<String, Object> contentSynchronisationNewValues = (Map<String, Object>) newValue.getValue();
                    for (Map.Entry<String, Object> contentSynchronisationNewValue :
                            contentSynchronisationNewValues.entrySet()) {
                        if ("enabled".equals(contentSynchronisationNewValue.getKey())) {
                            contentSynchronisation.setEnabled((Boolean) contentSynchronisationNewValue.getValue());
                        } else if ("properties".equals(contentSynchronisationNewValue.getKey())) {
                            PropertiesContent propertiesContent = new PropertiesContent();
                            if (contentSynchronisation.getProperties() != null) {
                                propertiesContent = contentSynchronisation.getProperties();
                            }
                            HashMap<String, Object> propertiesContentNewValues =
                                    (HashMap<String, Object>) contentSynchronisationNewValue.getValue();
                            if (!propertiesContentNewValues.isEmpty()) {
                                propertiesContent.setEnabled((Boolean) propertiesContentNewValues.get("enabled"));
                                contentSynchronisation.setProperties(propertiesContent);
                            }
                        } else if ("statistics".equals(contentSynchronisationNewValue.getKey())) {
                            StatisticsContent statisticsContent = new StatisticsContent();
                            if (contentSynchronisation.getStatistics() != null) {
                                statisticsContent = contentSynchronisation.getStatistics();
                            }
                            HashMap<String, Object> statisticsContentNewValues =
                                    (HashMap<String, Object>) contentSynchronisationNewValue.getValue();
                            if (!statisticsContentNewValues.isEmpty()) {
                                statisticsContent.setEnabled((Boolean) statisticsContentNewValues.get("enabled"));
                                contentSynchronisation.setStatistics(statisticsContent);
                            }
                        } else if ("source".equals(contentSynchronisationNewValue.getKey())) {
                            SourceContent sourceContent = new SourceContent();
                            if (contentSynchronisation.getSource() != null) {
                                sourceContent = contentSynchronisation.getSource();
                            }
                            HashMap<String, Object> sourceContentNewValues =
                                    (HashMap<String, Object>) contentSynchronisationNewValue.getValue();
                            if (!sourceContentNewValues.isEmpty()) {
                                sourceContent.setOriginAbsenceDetection(
                                        (Boolean) sourceContentNewValues.get("originAbsenceDetection"));
                                contentSynchronisation.setSource(sourceContent);
                            }
                        }
                    }
                }
            } catch (ClassCastException e) {
                throw new IllegalArgumentException("No values under " + newValue.getKey());
            }
        }
    }

    private static void addValueToPackageTypeConfigIfNeeded(String key, Map<String, String> packageTypeConfig,
            Map.Entry<String, Object> entry) {
        if (key.equals(entry.getKey())) {
            packageTypeConfig.put(key, String.valueOf(entry.getValue()));
        }
    }
}
