package org.artifactory.repo.service.validation;

import org.artifactory.api.security.AuthorizationException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.RepoDetailsType;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.Collection;
import java.util.Set;

import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;

/**
 * @author Mor Merhav
 */
@Component
public class ProjectAdminRepoConfigValidatorImpl implements ProjectAdminRepoConfigValidator {

    private final AuthorizationService authService;
    private static final Set<RepoDetailsType> ALLOWED_REPO_TYPES = Set.of(LOCAL, REMOTE, VIRTUAL, RELEASE_BUNDLE);

    @Autowired
    public ProjectAdminRepoConfigValidatorImpl(AuthorizationService authService) {
        this.authService = authService;
    }

    @Override
    public void validate(RepoConfig repoConfig) {
        if (repoConfig == null) {
            return;
        }
        validate(Set.of(repoConfig));
    }

    @Override
    public void validate(Collection<? extends RepoConfig> repoConfigs) {
        if (repoConfigs == null) {
            return;
        }

        if (!authService.isAnyProjectAdmin()) {
            throw new AuthorizationException("Only an admin user can save the artifactory configuration.");
        }
        if (authService.isAdmin()) {
            return;
        }

        boolean validationFailed = false;
        StringBuilder stringBuilder = null;
        for (RepoConfig repoConfig : repoConfigs) {
            if (!ALLOWED_REPO_TYPES.contains(repoConfig.getRepoType())) {
                validationFailed = true;
                if (stringBuilder == null) {
                    stringBuilder = new StringBuilder("Project admin is not allowed to create repositories: ").append(
                            repoConfig.getKey());
                } else {
                    stringBuilder.append(",").append(repoConfig.getKey());
                }
            }
        }

        if (validationFailed) {
            throw new AuthorizationException(stringBuilder.toString());
        }
    }
}
