package org.artifactory.security.access.util;

import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.security.ArtifactoryUserStatus;
import org.artifactory.security.SaltedPassword;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.UserMFAStatus;
import org.artifactory.security.access.AccessTokenAuthenticationException;
import org.artifactory.security.access.AccessTokenAuthenticationProvider;
import org.artifactory.security.access.ArtifactoryAdminScopeToken;
import org.artifactory.storage.db.security.service.access.UserMapper;
import org.artifactory.util.CollectionUtils;
import org.jfrog.access.client.AccessClientHttpException;
import org.jfrog.access.client.auth.UserPrincipal;
import org.jfrog.access.client.model.ErrorsModel;
import org.jfrog.access.client.model.MessageModel;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.model.MessageModelCode;
import org.jfrog.access.rest.user.UserWithGroups;
import org.jfrog.access.user.UsersUtil;
import org.springframework.security.authentication.BadCredentialsException;
import org.springframework.security.authentication.CredentialsExpiredException;
import org.springframework.security.authentication.DisabledException;
import org.springframework.security.authentication.LockedException;
import org.springframework.security.core.Authentication;

import java.util.ArrayList;
import java.util.Optional;
import java.util.Set;

import static org.artifactory.storage.db.security.service.access.GroupMapper.fromAccessRealm;


/**
 * @author Mor Merhav
 */
@Slf4j
@UtilityClass
public class AccessAuthProviderUtils {

    public final int SC_TOO_MANY_REQUESTS = 429;

    public void handleAccessClientHttpException(Authentication authentication, AccessClientHttpException e) {
        if (shouldHandleHttpException(e)) {
            int statusCode = e.getStatusCode();
            log.debug("{}, Access didn't authenticate user: '{}'", statusCode, authentication.getPrincipal());
            ErrorsModel errorsModel = e.getErrorsModel();
            if (errorsModel != null) {
                Optional<MessageModel> tooManyRequests = isCode(errorsModel, MessageModelCode.CODE_TOO_MANY_REQUESTS);
                Optional<MessageModel> passwordExpired = isCode(errorsModel, MessageModelCode.CODE_PASSWORD_EXPIRED);
                Optional<MessageModel> userDisabled = isCode(errorsModel, MessageModelCode.CODE_DISABLED);
                Optional<MessageModel> tokenFailed =
                        isCode(errorsModel, MessageModelCode.CODE_TOKEN_VERIFICATION_FAILED);
                if (tooManyRequests.isPresent()) {
                    throw new LockedException(tooManyRequests
                            .map(MessageModel::getMessage).orElse("Too many attempts, please try again later"), e);
                }
                if (passwordExpired.isPresent()) {
                    throw new CredentialsExpiredException("Password has expired", e);
                }
                if (userDisabled.isPresent()) {
                    throw new DisabledException("User is disabled", e);
                }
                if (tokenFailed.isPresent()) {
                    throw new AccessTokenAuthenticationException(
                            "Token failed verification: " +
                                    tokenFailed.map(MessageModel::getMessage).orElse("invalid"));
                }
            }
            throw new BadCredentialsException("Bad credentials");
        }
    }

    private boolean shouldHandleHttpException(AccessClientHttpException e) {
        int statusCode = e.getStatusCode();
        return statusCode < (HttpStatus.SC_NOT_FOUND) || statusCode == SC_TOO_MANY_REQUESTS;
    }

    private Optional<MessageModel> isCode(ErrorsModel errorsModel, String code) {
        return errorsModel.getErrors().stream().filter(model -> code.equals(model.getCode())).findFirst();
    }

    public SimpleUser toArtifactoryUser(UserPrincipal userPrincipal, String password) {
        return toArtifactoryUser(userPrincipal, password, null);
    }

    public SimpleUser toArtifactoryUser(UserPrincipal userPrincipal, String password, ServiceId serviceId) {
        UserWithGroups user = userPrincipal.user();
        Set<UsersUtil.BuiltInUserProperty> userProperties = userPrincipal.userProperties();
        Set<String> tokenScopes = userPrincipal.tokenScopes();
        UserInfoBuilder builder = new UserInfoBuilder(user.getUsername())
                .email(StringUtils.isBlank(user.getEmail()) ? null : user.getEmail())
                .admin(isAdmin(userProperties, tokenScopes, serviceId))
                .policyViewer(userProperties.contains(UsersUtil.BuiltInUserProperty.policy_viewer))
                .policyManager(userProperties.contains(UsersUtil.BuiltInUserProperty.policy_manager))
                .watchManager(userProperties.contains(UsersUtil.BuiltInUserProperty.watch_manager))
                .reportsManager(userProperties.contains(UsersUtil.BuiltInUserProperty.reports_manager))
                .updatableProfile(userProperties.contains(UsersUtil.BuiltInUserProperty.updatable_profile))
                .passwordDisabled(userProperties.contains(UsersUtil.BuiltInUserProperty.disabled_password))
                .lastLogin(user.getLastLoginTime(), user.getLastLoginIp())
                .credentialsExpired(user.isPasswordExpired())
                .groupAdmin(null) // This is checked later on via UserImpl#isEffectiveAdmin
                .groups(UserMapper.groupsFromUser(user))
                .password(new SaltedPassword(password, null))
                .status(ArtifactoryUserStatus.getByValue(user.getStatus().getName()))
                .mfaStatus(UserMFAStatus.valueOf(user.getMfaStatus().toString()))
                .realm(fromAccessRealm(user.getRealm()));

        return new SimpleUser(builder.build());
    }

    private boolean isAdmin(Set<UsersUtil.BuiltInUserProperty> userProperties, Set<String> scopes,
            ServiceId serviceId) {
        if (CollectionUtils.notNullOrEmpty(scopes) && serviceId != null) {
            boolean scopeAdmin = isScopeAdmin(scopes, serviceId);
            if (AccessTokenAuthenticationProvider.isIdentityToken(new ArrayList<>(scopes))) {
                return userProperties.contains(UsersUtil.BuiltInUserProperty.artifactory_admin) || scopeAdmin;
            }
            return scopeAdmin;
        }
        return userProperties.contains(UsersUtil.BuiltInUserProperty.artifactory_admin);
    }

    public boolean isScopeAdmin(Set<String> scopes, ServiceId serviceId) {
        return scopes.stream()
                .filter(ArtifactoryAdminScopeToken::accepts)
                .anyMatch(scope -> ArtifactoryAdminScopeToken.isAdminScopeOnService(scope, serviceId));
    }
}
