package org.artifactory.security.utils;

import org.apache.commons.lang3.StringUtils;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.jfrog.access.common.ServiceType;
import org.springframework.security.core.Authentication;

import static org.artifactory.addon.xray.XrayAddon.ARTIFACTORY_XRAY_USER;

/**
 * @author Alexei Vainshtein
 */
public class AuthenticationUtils {

    private AuthenticationUtils() {
    }

    /**
     * Returns if the user is an xray user/token
     *
     * @param currentUser    - The user that performed the request.
     * @param authentication - The authentication object
     * @return true if the user/authentication is from Xray, false otherwise.
     */
    public static boolean isXrayAuthenticationUser(String currentUser, Authentication authentication) {
        return ARTIFACTORY_XRAY_USER.equals(currentUser) ||
                isXrayTokenAuthentication(authentication);
    }

    /**
     * Returns if the user is an xray token
     *
     * @param authentication - The authentication object
     * @return true if the authentication is Xray token, false otherwise
     */
    static boolean isXrayTokenAuthentication(Authentication authentication) {
        if (authentication instanceof AccessTokenAuthentication accessToken) {
            String subject = accessToken.getAccessToken().getSubject();
            if (StringUtils.isBlank(subject)) {
                return false;
            }

            return subject.startsWith(ServiceType.XRAY + "@")
                    || subject.startsWith(ServiceType.XRAY_INDEXER + "@")
                    || subject.startsWith(ServiceType.XRAY_ANALYSIS + "@")
                    || subject.startsWith(ServiceType.XRAY_PERSIST + "@");

        }
        return false;
    }

    public static boolean isPipelinesAuthentication(Authentication authentication) {
        if (authentication instanceof AccessTokenAuthentication) {
            String issuer = ((AccessTokenAuthentication) authentication).getAccessToken().getIssuer();
            if (StringUtils.isBlank(issuer)) {
                return false;
            }

            return issuer.startsWith("jfpip@");
        }
        return false;
    }
}
