package org.artifactory.util;

import lombok.experimental.UtilityClass;
import org.apache.commons.collections.MapUtils;
import org.apache.http.HttpStatus;
import org.apache.http.NoHttpResponseException;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.conn.HttpHostConnectException;
import org.apache.http.impl.client.CloseableHttpClient;
import org.artifactory.api.rest.constant.RestConstants;
import org.artifactory.api.rest.constant.SystemRestConstants;
import org.artifactory.repo.http.CloseableHttpExecute;
import org.artifactory.rest.exception.NotFoundException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.io.UncheckedIOException;
import java.util.Map;
import java.util.function.Function;

/**
 * Utility class to get the Artifactory service version via the new API with backward compatibility for older versions.
 *
 * @author Andrey Tashlitsky
 */
@UtilityClass
public class ArtifactoryVersionUtil {

    private static final Logger log = LoggerFactory.getLogger(ArtifactoryVersionUtil.class);


    public CloseableHttpResponse getArtifactoryServiceVersion(CloseableHttpExecute closeableHttpExecute,
            String instanceUrl, Map<String, String> headers) {
        return getArtifactoryServiceVersionFunction(httpGet -> {
            try {
                return closeableHttpExecute.executeMethod(httpGet);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }, instanceUrl, headers);
    }

    public CloseableHttpResponse getArtifactoryServiceVersion(CloseableHttpClient closeableHttpClient,
            String instanceUrl, Map<String, String> headers) {
        return getArtifactoryServiceVersionFunction(httpGet -> {
            try {
                return closeableHttpClient.execute(httpGet);
            } catch (NoHttpResponseException | HttpHostConnectException e) {
                throw new NotFoundException(String.format("Resource not found on remote url: '%s'", instanceUrl), e);
            } catch (IOException e) {
                throw new UncheckedIOException(e);
            }
        }, instanceUrl, headers);
    }

    private CloseableHttpResponse getArtifactoryServiceVersionFunction(
            Function<HttpGet, CloseableHttpResponse> getServiceVersion, String instanceUrl,
            Map<String, String> headers) {
        String basePath = instanceUrl + "/" + RestConstants.PATH_API + "/" + SystemRestConstants.PATH_ROOT + "/";
        HttpGet serviceVersionHttpGet = new HttpGet(basePath + SystemRestConstants.PATH_VERSION_SERVICE);
        setHeaders(serviceVersionHttpGet, headers);
        return getServiceVersionWithBackwardCompatibility(getServiceVersion, headers, basePath, serviceVersionHttpGet);
    }

    CloseableHttpResponse getServiceVersionWithBackwardCompatibility(
            Function<HttpGet, CloseableHttpResponse> getServiceVersion, Map<String, String> headers, String basePath,
            HttpGet serviceVersionHttpGet) {
        CloseableHttpResponse versionResponse = getServiceVersion.apply(serviceVersionHttpGet);

        if (versionResponse.getStatusLine().getStatusCode() == HttpStatus.SC_NOT_FOUND) {
            try {
                versionResponse.close();
            } catch (IOException e) {
                log.error("Unable to close connection: {} ", e.getMessage());
                log.debug("", e);
            }
            HttpGet oldVersionGet = new HttpGet(basePath + SystemRestConstants.PATH_VERSION);
            setHeaders(serviceVersionHttpGet, headers);
            return getServiceVersion.apply(oldVersionGet);
        }
        return versionResponse;
    }

    private void setHeaders(HttpGet httpGet, Map<String, String> headers) {
        if (MapUtils.isEmpty(headers)) {
            return;
        }
        for (Map.Entry<String, String> header : headers.entrySet()) {
            httpGet.addHeader(header.getKey(), header.getValue());
        }
    }
}
