package org.artifactory.engine;

import org.apache.http.HttpHeaders;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.worker.WorkersAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.request.InternalRequestContext;
import org.artifactory.request.RequestClientType;
import org.artifactory.request.RequestThreadLocal;
import org.artifactory.request.RequestWrapper;
import org.artifactory.resource.RepoResourceInfo;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.storage.binstore.common.BinaryElementRequestImpl;
import org.jfrog.storage.binstore.ifc.ManualRedirectType;
import org.jfrog.storage.binstore.ifc.RedirectType;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;

import javax.servlet.http.HttpServletRequest;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.stream.Stream;

import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.MANUAL_REDIRECT_TYPE;
import static org.jfrog.storage.binstore.ifc.model.BinaryElementHeaders.SKIP_CONTENT_DISPOSITION;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Shay Bagants
 */
public class DownloadServiceImplTest extends ArtifactoryHomeBoundJUnitTest {

    private DownloadServiceImpl downloadService;

    @Mock
    private InternalRequestContext requestContext;
    @Mock
    private ArtifactoryRequest request;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private DownloadRedirectHelper downloadRedirectHelper;
    @Mock
    private ArtifactoryBinaryService binaryService;
    @Mock
    private CentralConfigService centralConfig;
    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private AolService aolService;
    @Mock
    private WorkersAddon workersAddon;
    private AutoCloseable mocks;

    private static Stream<Arguments> provideRequestData() {
        return Stream.of(
                Arguments.of("xray/1.0", "xray", "xray"),
                Arguments.of("xray/1.0", "test-user", null),
                Arguments.of("jfrog-cli", "xray", null),
                Arguments.of("xray/1.0", "token:jfrt@123", null),
                Arguments.of("jfrog-cli", "token:jfxr@123", null),
                Arguments.of("x-ray/1.0", "token:jfxr@123", null)
        );
    }

    private static Stream<Arguments> provideRequestDataForAccessTokenAuthentication() {
        return Stream.of(
                //agent, user, expectedResult
                Arguments.of("xray/1.0", "xray", null),
                Arguments.of("xray/1.0", "test-user", null),
                Arguments.of("jfrog-cli", "xray", null),
                Arguments.of("xray/1.0", "token:jfrt@123", null),
                Arguments.of("jfrog-cli", "jfxr@123", null),
                Arguments.of("xray/1.0", "jfxr@123", "jfxr@123"),
                Arguments.of("x-ray/1.0", "jfxidx@123", null),
                Arguments.of("xray/1.0", "jfxidx@123", "jfxidx@123"),
                Arguments.of("x-ray/1.0", "jfxpst@123", null),
                Arguments.of("xray/1.0", "jfxpst@123", "jfxpst@123"),
                Arguments.of("x-ray/1.0", "jfxana@123", null),
                Arguments.of("xray/1.0", "jfxana@123", "jfxana@123")
        );
    }

    private static Stream<Arguments> getCdnConfigurationsData() {
        return Stream.of(
                //isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, manualSentType, expectedType
                //AOL
                Arguments.of(true, true, false, false, true, false, ManualRedirectType.CF, RedirectType.CLOUD_FRONT),
                Arguments.of(true, true, false, false, true, false, ManualRedirectType.DEFAULT,
                        RedirectType.CLOUD_FRONT),
                Arguments.of(true, false, true, false, true, false, ManualRedirectType.CF, RedirectType.CLOUD_FRONT),
                Arguments.of(true, true, false, false, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT),
                Arguments.of(true, false, true, false, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT),
                Arguments.of(true, false, false, true, true, true, ManualRedirectType.CF, RedirectType.CLOUD_FRONT),
                Arguments.of(true, false, false, true, true, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(true, false, false, true, true, false, ManualRedirectType.DEFAULT, RedirectType.S3),
                Arguments.of(true, false, false, false, true, true, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(true, false, false, false, true, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(false, true, false, false, true, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(false, false, true, false, true, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(false, false, false, false, true, false, ManualRedirectType.S3, RedirectType.S3),
                //OnPrem - isCdn cannot be true onPrem (as of 29/03/20)
                Arguments.of(false, true, false, false, false, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(false, false, true, false, false, false, ManualRedirectType.S3, RedirectType.S3),
                Arguments.of(false, false, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT),
                Arguments.of(false, false, false, true, false, true, ManualRedirectType.NONE, RedirectType.DIRECT),
                Arguments.of(false, false, false, true, false, true, ManualRedirectType.DEFAULT, RedirectType.DIRECT),
                Arguments.of(false, false, false, true, false, false, ManualRedirectType.NONE, RedirectType.DIRECT),
                //Manual DIRECT request - will return direct when it's not AOL
                Arguments.of(false, true, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT),
                Arguments.of(false, true, false, false, false, false, ManualRedirectType.NONE, RedirectType.DIRECT),
                //Manual DIRECT request - will ignore the manual request when it's AOL
                Arguments.of(true, true, false, false, true, false, ManualRedirectType.NONE, RedirectType.CLOUD_FRONT),
                Arguments.of(false, false, true, false, true, false, ManualRedirectType.NONE, RedirectType.S3)
        );
    }

    private static Stream<Arguments> getCdnConfigurationsWithCdnListData() {
        return Stream.of(
                //isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, manualSentType, expectedType
                //if list of allowed cdn repos configured isCdn is ignored
                Arguments.of(false, true, false, false, true, true, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                Arguments.of(false, true, false, false, true, false, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                Arguments.of(false, false, true, false, true, true, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                Arguments.of(false, false, true, false, true, false, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                Arguments.of(false, false, false, true, true, true, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                Arguments.of(false, false, false, true, true, false, ManualRedirectType.S3, RedirectType.CLOUD_FRONT),
                //checking that for non-AOL list of allowed CDN repos are not applied
                Arguments.of(false, true, false, false, false, false, ManualRedirectType.S3, RedirectType.S3)
        );
    }

    private static Stream<Arguments> threadLocalDataProvider() {
        return Stream.of(
                Arguments.of(ManualRedirectType.S3.getType(), ManualRedirectType.S3, false),
                Arguments.of(ManualRedirectType.CF.getType(), ManualRedirectType.CF, true),
                Arguments.of(ManualRedirectType.CF.getType(), ManualRedirectType.DEFAULT, false),
                Arguments.of(ManualRedirectType.NONE.getType(), ManualRedirectType.NONE, false),
                Arguments.of(null, ManualRedirectType.DEFAULT, false)
        );
    }

    @BeforeEach
    public void setup() {
        downloadService = spy(new DownloadServiceImpl());
        mocks = MockitoAnnotations.openMocks(this);
        when(requestContext.getRequest()).thenReturn(request);
        reset(downloadService);
        downloadService.authorizationService = authorizationService;
        downloadService.downloadRedirectHelper = downloadRedirectHelper;
        downloadService.addonsManager = addonsManager;
        downloadService.binaryService = binaryService;
        downloadService.centralConfig = centralConfig;
        downloadService.repositoryService = repositoryService;

        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterEach
    public void unbind() throws Exception {
        mocks.close();
        unbindArtifactoryHome();
        RequestThreadLocal.unbind();
        ArtifactoryContextThreadBinder.unbind();
    }

    @ParameterizedTest
    @MethodSource("threadLocalDataProvider")
    void testGetManualRedirectType(String threadHeader, ManualRedirectType expected, boolean isAol) {
        bindRequest(threadHeader);
        Enumeration enumeration = mock(Enumeration.class);
        when(enumeration.hasMoreElements()).thenReturn(false);
        when(request.getHeaderValues(MANUAL_REDIRECT_TYPE.getHeaderName())).thenReturn(enumeration);
        doReturn(isAol).when(downloadService).canUseCdnWithLicense(any(Repo.class));
        ManualRedirectType type = downloadService.getManualRedirectType(requestContext, mock(Repo.class));
        assertEquals(expected, type);
    }

    @ParameterizedTest
    @MethodSource("provideRequestData")
    void testGetXrayUserName(String userAgent, String principal, String expectedResult) {
        when(authorizationService.currentUsername()).thenReturn(principal);
        when(request.getHeader(HttpHeaders.USER_AGENT)).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, null);
        assertEquals(result, expectedResult);
    }

    @ParameterizedTest
    @MethodSource("provideRequestDataForAccessTokenAuthentication")
    void testGetXrayUserNameUsingAccessTokenAuthentication(String userAgent, String principal,
            String expectedResult) {
        AccessTokenAuthentication accessTokenAuthentication = mock(AccessTokenAuthentication.class);
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(accessTokenAuthentication.getAccessToken()).thenReturn(jwtAccessToken);
        when(jwtAccessToken.getSubject()).thenReturn(principal);
        when(request.getHeader(HttpHeaders.USER_AGENT)).thenReturn(userAgent);
        String result = downloadService.getXrayUserName(requestContext, accessTokenAuthentication);
        Assert.assertEquals(result, expectedResult);
    }

    @ParameterizedTest
    @MethodSource("getCdnConfigurationsData")
    void testGetRedirectTypeForRepo_differentCdnConfigurations(boolean isCdn, boolean isEplus, boolean isEdge,
            boolean isEnterprise, boolean isAol, boolean isAnonymous, ManualRedirectType typeSent,
            RedirectType expectedType) {
        setupCdn(false);
        testDifferentCdnConfigurations(isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, typeSent,
                expectedType);
    }

    @ParameterizedTest
    @MethodSource("getCdnConfigurationsWithCdnListData")
    void testGetRedirectTypeForRepo_differentCdnConfigurationsWithCdnList(boolean isCdn, boolean isEplus,
            boolean isEdge, boolean isEnterprise, boolean isAol, boolean isAnonymous, ManualRedirectType typeSent,
            RedirectType expectedType) {
        setupCdn(true);
        testDifferentCdnConfigurations(isCdn, isEplus, isEdge, isEnterprise, isAol, isAnonymous, typeSent,
                expectedType);
    }

    private void testDifferentCdnConfigurations(boolean isCdn, boolean isEplus, boolean isEdge,
            boolean isEnterprise, boolean isAol, boolean isAnonymous, ManualRedirectType typeSent,
            RedirectType expectedType) {
        RepoResource resource = mock(RepoResource.class);
        Repo repo = mock(Repo.class);
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);

        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(repo.getDescriptor()).thenReturn(repoDescriptor);

        // using params
        when(repoDescriptor.isCdnRedirect()).thenReturn(isCdn);
        when(addonsManager.isEntitled(EntitlementType.CDN)).thenReturn(isEplus || isEdge);
        when(addonsManager.isEntitled(EntitlementType.ANONYMOUS_CDN)).thenReturn(isEplus || isEnterprise || isEdge);
        when(addonsManager.isEntitled(EntitlementType.BINARY_PROVIDER_DOWNLOAD_REDIRECT)).thenReturn(isEdge || isEplus);
        when(aolService.isAol()).thenReturn(isAol);
        when(authorizationService.isAnonymous()).thenReturn(isAnonymous);

        RedirectType redirectType = downloadService.getRedirectTypeForRepo(requestContext, repo, resource, typeSent);

        assertEquals(expectedType, redirectType);
    }

    @Test
    void testGetSignUrl() {
        // setup
        RepoResource resource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);
        Map<String, String> headers = downloadService
                .getHeaderMapForRedirect(repoResourceInfo, "user", RedirectType.S3, RequestClientType.DEBIAN,
                        null, ManualRedirectType.DEFAULT, ProjectsService.DEFAULT_PROJECT,
                        "local", "npm");
        BinaryElementRequest request = new BinaryElementRequestImpl(resource.getInfo().getSha1(),null,null,0L, headers);
        when(binaryService.getSignedUrl(request)).thenReturn("http://signed.url");

        // test
        String url = downloadService
                .getSignedUrl(resource, "user", RedirectType.S3, RequestClientType.DEBIAN, null,
                        ManualRedirectType.DEFAULT, ProjectsService.DEFAULT_PROJECT, "local", "npm");

        // verify
        assertEquals("http://signed.url", url);
    }

    @Test
    void testGetHeaderMapForRedirect() {
        // setup
        RepoPath repoPath = mock(RepoPath.class);
        when(repoPath.getName()).thenReturn("repoName");
        when(repoPath.getPath()).thenReturn("repoPath");

        RepoResourceInfo resourceInfo = mock(RepoResourceInfo.class);
        when(resourceInfo.getRepoPath()).thenReturn(repoPath);

        // test
        Map<String, String> headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.UNKNOWN,
                        null, ManualRedirectType.DEFAULT, ProjectsService.DEFAULT_PROJECT,
                        "local", "npm");

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "false");

        // test
        headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.DEBIAN,
                        null, ManualRedirectType.DEFAULT, ProjectsService.DEFAULT_PROJECT,
                        "local", "npm");

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "true");

        // test
        headers = downloadService
                .getHeaderMapForRedirect(resourceInfo, "user", RedirectType.S3, RequestClientType.APACHE_IVY,
                        null, ManualRedirectType.DEFAULT, ProjectsService.DEFAULT_PROJECT,
                        "local", "npm");

        // verify
        Assert.assertEquals(headers.get(SKIP_CONTENT_DISPOSITION.getHeaderName()), "false");
    }

    @ParameterizedTest
    @MethodSource("shouldRedirectProvider")
    void testShouldRedirect(boolean headRequest, boolean expectedResult) {
        LocalRepoDescriptor repoDescriptor = mock(LocalRepoDescriptor.class);
        when(repoDescriptor.isReal()).thenReturn(true);
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(request.isHeadOnly()).thenReturn(headRequest);
        RepoResource resource = mock(RepoResource.class);
        RepoResourceInfo repoResourceInfo = mock(RepoResourceInfo.class);
        when(repoResourceInfo.getSha1()).thenReturn("cf23df2207d99a74fbe169e3eba035e633b65d94");
        when(resource.getInfo()).thenReturn(repoResourceInfo);

        boolean shouldRedirect = downloadService.shouldRedirect(requestContext, null, resource, repoDescriptor, true);
        assertEquals(shouldRedirect, expectedResult);
    }

    @Test
    void testOnContextLoadedNotifyRepoRedirectAllowance() {
        // setup
        LocalRepoConfig gen1 = LocalRepoConfig.create("gen1", RepoType.Generic);
        FederatedRepoConfig gen2 = FederatedRepoConfig.create("gen2", RepoType.Generic);
        when(repositoryService.getLocalRepoConfigsIncludingInfoRepos()).thenReturn(List.of(gen1));
        when(repositoryService.getFederatedRepoConfigs()).thenReturn(List.of(gen2));

        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(0)).logWarnRedirectConfigured();

        // setup
        gen1.getRepoTypeConfig().setDownloadRedirect(true);
        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(1)).logWarnRedirectConfigured();
        clearInvocations(downloadService);

        // setup
        gen2.getRepoTypeConfig().setDownloadRedirect(true);
        // test
        downloadService.onLicenseLoaded();
        // verify non have redirect enabled by default
        verify(downloadService, times(1)).logWarnRedirectConfigured();
        clearInvocations(downloadService);
    }

    public static Stream<Arguments> shouldRedirectProvider() {
        return Stream.of(
                Arguments.of(true, false),
                Arguments.of(false, true)
        );
    }

    private void bindRequest(String header) {
        RequestWrapper requestWrapper = mock(RequestWrapper.class);
        HttpServletRequest request = mock(HttpServletRequest.class);
        when(requestWrapper.getRequest()).thenReturn(request);
        when(request.getHeader(MANUAL_REDIRECT_TYPE.getHeaderName())).thenReturn(header);
        RequestThreadLocal.bind(requestWrapper);
    }

    private void setupCdn(boolean withCdnList) {
        if (withCdnList) {
            when(downloadRedirectHelper.isCdnForcedForThisRepoType(any())).thenReturn(true);
        }
        when(downloadRedirectHelper.isRedirectEnable(any(), any())).thenReturn(true);
        when(downloadRedirectHelper.isArtifactGreaterEqualsRedirectThreshold(any())).thenReturn(true);
        when(binaryService.canCloudProviderGenerateRedirection(any())).thenReturn(true);
        when(aolService.isAol()).thenReturn(true);
    }
}