package org.artifactory.event;

import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.event.migration.EventTableRepoKeyMigration;
import org.artifactory.schedule.TaskService;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorStatus;
import org.artifactory.storage.db.event.entity.enums.NodeEventCursorType;
import org.artifactory.storage.db.event.error.service.NodeEventsErrorsStorageService;
import org.artifactory.storage.db.event.model.NodeEventCursor;
import org.artifactory.storage.db.event.service.NodeEventCursorService;
import org.artifactory.storage.event.EventsService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.joda.time.DateTime;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.quartz.JobExecutionException;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Date;
import java.util.Iterator;
import java.util.List;

import static java.util.concurrent.TimeUnit.HOURS;
import static org.artifactory.common.ConstantValues.ageOfEventsLogEntriesToDiscardDays;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.anyLong;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@Test
public class EventsLogCleanUpServiceTest extends ArtifactoryHomeBoundTest {

    @Mock
    private EventsService eventsService;

    @Mock
    private NodeEventsErrorsStorageService eventsErrorsService;

    @Mock
    private TaskService taskService;

    @Mock
    NodeEventCursorService nodeEventCursorService;
    @Mock
    EventTableRepoKeyMigration eventTableRepoKeyMigration;

    private EventsLogCleanUpService eventsLogCleanUpService;

    @BeforeMethod
    public void init() {
        MockitoAnnotations.openMocks(this);
        eventsLogCleanUpService = Mockito.spy(new EventsLogCleanUpService(taskService, eventsService,
                nodeEventCursorService, eventsErrorsService,eventTableRepoKeyMigration));
        ArtifactoryHome
                .get().getArtifactoryProperties()
                .setProperty(ageOfEventsLogEntriesToDiscardDays.getPropertyName(), "0");
    }

    @Test(dataProvider = "MaxBoundaryProvider")
    public void testGetMaxBoundary(long minCursor, long minError, long configBoundary,
            EventsLogCleanUpService.MaxBoundary expected) {
        // Given
        doAnswer(invocation -> configBoundary).when(eventsLogCleanUpService).getConfigBoundary(any(Date.class));
        when(nodeEventCursorService.oldestTimestampForEventLogOperator()).thenReturn(minCursor);
        when(eventsErrorsService.getFirstEventErrorTimestampWithEventId()).thenReturn(minError);

        // When
        EventsLogCleanUpService.MaxBoundary maxBoundary = eventsLogCleanUpService.getMaxBoundary();

        // Then
        Assert.assertEquals(maxBoundary.getTimestamp(), expected.getTimestamp());
        Assert.assertEquals(maxBoundary.isShouldDeleteExpiredEvents(), expected.isShouldDeleteExpiredEvents());
        verify(eventsLogCleanUpService, times(1)).getMaxBoundary();
        verify(nodeEventCursorService, times(1)).oldestTimestampForEventLogOperator();
        verify(eventsErrorsService, times(1)).getFirstEventErrorTimestampWithEventId();

    }

    @DataProvider(name = "MaxBoundaryProvider")
    public Object[][] getMaxBoundaryProvider() {
        long time = System.currentTimeMillis();

        // {minCursor, minError, configBoundary, expected}
        return new Object[][] {
                {time - HOURS.toMillis(10),
                        time - HOURS.toMillis(5),
                        time - HOURS.toMillis(20),
                        EventsLogCleanUpService.MaxBoundary.builder()
                                .timestamp(time - HOURS.toMillis(10)).build()},
                {time - HOURS.toMillis(10),
                        time - HOURS.toMillis(15),
                        time - HOURS.toMillis(20),
                        EventsLogCleanUpService.MaxBoundary.builder()
                                .timestamp(time - HOURS.toMillis(15)).build()},
                {time - HOURS.toMillis(10),
                        time - HOURS.toMillis(5),
                        time,
                        EventsLogCleanUpService.MaxBoundary.builder()
                                .timestamp(time).shouldDeleteExpiredEvents(true).build()},
                {time - HOURS.toMillis(10),
                        time - HOURS.toMillis(15),
                        time,
                        EventsLogCleanUpService.MaxBoundary.builder()
                                .timestamp(time).shouldDeleteExpiredEvents(true).build()}
        };
    }


    @Test
    public void testCleanUpLargeRange() throws JobExecutionException {
        when(eventsService.getFirstEventTimestamp()).thenReturn(new Date().getTime() - HOURS.toMillis(10));
        when(eventsService.getEventsCount()).thenReturn(10L);
        eventsLogCleanUpService.cleanup();
        verify(eventsService, atLeast(3)).deleteRange(anyLong(), anyLong());
        verify(eventsErrorsService, atLeast(1)).deleteBeforeWithEventId(anyLong());
    }

    @Test
    public void testCleanUpLargeRangeOldestMarkerIsFromBeforeAllEvents() throws JobExecutionException {
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ageOfEventsLogEntriesToDiscardDays.getPropertyName(), "1");
        when(eventsService.getFirstEventTimestamp()).thenReturn(new Date().getTime() - HOURS.toMillis(10));
        when(nodeEventCursorService.oldestTimestampForEventLogOperator())
                .thenReturn((new Date().getTime() - HOURS.toMillis(11)));
        when(eventsErrorsService.getFirstEventErrorTimestampWithEventId())
                .thenReturn((new Date().getTime() - HOURS.toMillis(10)));
        when(eventsService.getEventsCount()).thenReturn(10L);
        eventsLogCleanUpService.cleanup();
        verify(eventsService, times(0)).deleteRange(anyLong(), anyLong());
        verify(eventsErrorsService, times(0)).deleteBeforeWithEventId(anyLong());
    }

    @Test
    public void testCleanUpSmallRange() throws JobExecutionException {

        // Given
        DateTime time = new DateTime(100000);
        /*
          cur time         = 100000
          conf boundary    = 100000
          first            = 100000 - 1H
          events           = 0
          errors           = 0
         */

        doAnswer(invocation -> {
            // Force the getConfigBoundary to use the same base time (to enable debug)
            ((Date) invocation.getArguments()[0]).setTime(time.getMillis());
            return invocation.callRealMethod();
        }).when(eventsLogCleanUpService).getConfigBoundary(any(Date.class));

        when(eventsService.getFirstEventTimestamp()).thenReturn(time.minusHours(1).getMillis());
        when(eventsService.getEventsCount()).thenReturn(10L);


        // When
        eventsLogCleanUpService.cleanup();

        // Then
        // Verify one time delete
        verify(eventsService, times(1)).deleteRange(anyLong(), anyLong());
        verify(eventsErrorsService, times(1)).deleteBeforeWithEventId(anyLong());
    }

    @Test
    public void testCleanUpAllEventsAreTooYoung() throws JobExecutionException {

        // Given
        DateTime time = new DateTime(1000);
        /*
          cur time         = 1000
          conf boundary    = 1000
          first            = 2000
          events           = 0
          errors           = 0
         */

        doAnswer(invocation -> {
            // Force the getConfigBoundary to use the same base time (to enable debug)
            ((Date) invocation.getArguments()[0]).setTime(time.getMillis());
            return invocation.callRealMethod();
        }).when(eventsLogCleanUpService).getConfigBoundary(any(Date.class));
        when(eventsService.getFirstEventTimestamp()).thenReturn(time.plusMillis(1000).getMillis());
        when(eventsService.getEventsCount()).thenReturn(10L);

        // When
        eventsLogCleanUpService.cleanup();

        // Then
        // Verify no cleanup
        verify(eventsService, never()).deleteRange(anyLong(), anyLong());
        verify(eventsErrorsService, never()).deleteBeforeWithEventId(anyLong());
    }

    @Test
    public void testMarkFederatedAsOOS() {
        when(nodeEventCursorService.getEventIdCursorsBy(
                NodeEventCursorStatus.SYNC, 3L))
                .thenReturn(getNodeCursorData()
                );
        when(eventsService.countEventsRangeByRepoKey(0L, 3L, "repo2")).thenReturn(1);
        when(eventsService.countEventsRangeByRepoKey(0L, 3L, "repo")).thenReturn(0);
        eventsLogCleanUpService.markCursorsAsOutOfSyncIfNeeded(1L, 3L);

        Iterator<NodeEventCursor> cursorIter = getNodeCursorData().iterator();
        verify(nodeEventCursorService, times(1))
                .updateCursorWithStatus(cursorIter.next().getOperatorId(), NodeEventCursorStatus.OUT_OF_SYNC);
        verify(nodeEventCursorService, times(0))
                .updateCursorWithStatus(cursorIter.next().getOperatorId(), NodeEventCursorStatus.OUT_OF_SYNC);
    }

    @Test
    public void testGetEventsDeleteInterval() {
        long value;
        // check defaulkt value of 4 hours
        value = eventsLogCleanUpService.getEventsDeleteInterval();
        Assert.assertEquals(4 * 3600 * 1000, value);

        ArtifactoryHome
                .get().getArtifactoryProperties()
                .setProperty(ConstantValues.eventsCleanupIntervalHours.getPropertyName(), "1");
        value = eventsLogCleanUpService.getEventsDeleteInterval();
        Assert.assertEquals(1 * 3600 * 1000, value);

        // check fractions of an hour is supported
        ArtifactoryHome
                .get().getArtifactoryProperties()
                .setProperty(ConstantValues.eventsCleanupIntervalHours.getPropertyName(), "0.5");
        value = eventsLogCleanUpService.getEventsDeleteInterval();
        Assert.assertEquals( 1800 * 1000, value);

        // Invalid value (non-numeric) would fall back to default of 4 hours.
        ArtifactoryHome
                .get().getArtifactoryProperties()
                .setProperty(ConstantValues.eventsCleanupIntervalHours.getPropertyName(), "blabla");
        value = eventsLogCleanUpService.getEventsDeleteInterval();
        Assert.assertEquals( 4 * 3600 * 1000, value);
    }

    private List<NodeEventCursor> getNodeCursorData() {
        return List.of(NodeEventCursor.builder()
                        .operatorId("repo2-federated_replicationhttp://0.0.0.0:8081/artifactory/repo2-cursor")
                        .type(NodeEventCursorType.FEDERATED_REPLICATION).build(),
                NodeEventCursor.builder()
                        .operatorId("repo-federated_replicationhttp://0.0.0.0:8081/artifactory/repo-cursor")
                        .type(NodeEventCursorType.FEDERATED_REPLICATION).build());
    }
}