package org.artifactory.event.work;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.event.EventType;
import org.artifactory.event.priority.service.NodeEventPriorityService;
import org.artifactory.event.priority.service.model.PrioritizedNodeEvent;
import org.artifactory.opentracing.ArtifactoryJaegerTracerConfiguration;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.jfrog.common.ClockUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.timeout;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertNotEquals;
import static org.testng.Assert.assertNotNull;

/**
 * @author Uriah Levy
 */
@Test
public class EventOperatorDispatcherTest {

    @Mock
    NodeEventPriorityService priorityService;

    @Mock
    NodeEventOperator operator;

    @Mock
    HaCommonAddon haCommonAddon;

    private ArtifactoryTracer artifactoryTracer = new ArtifactoryTracer(
            ArtifactoryJaegerTracerConfiguration.initTracer());

    @Mock
    ArtifactoryContext artifactoryContext;

    @Mock
    AddonsManager addonsManager;

    @Mock
    HaAddon haAddon;

    @BeforeMethod
    public void beforeMethod() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        when(artifactoryContext.beanForType(ArtifactoryTracer.class)).thenReturn(artifactoryTracer);
        when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(haAddon.isHaEnabled()).thenReturn(true);
        when(priorityService.isDistributed()).thenReturn(true);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        artifactoryTracer.startSpanOnCurrentThread("Dispatcher Test");
    }

    @AfterMethod
    public void unbind() {
        ArtifactoryContextThreadBinder.unbind();
    }

    public void testRun() {
        when(operator.getOperatorId()).thenReturn("metadata-event-operator");
        when(operator.getNumberOfEventThreads()).thenReturn(1);
        when(operator.isOperatorEnabled()).thenReturn(true);

        PrioritizedNodeEvent prioritizedNodeEvent = new PrioritizedNodeEvent(123L, "a/b/c", EventType.props,
                "metadata-event-operator", 0, ClockUtils.epochMillis(), 1, PrioritizedNodeEvent.EventStatus.PENDING);
        final List<PrioritizedNodeEvent> nodeEvents = Collections.singletonList(prioritizedNodeEvent);
        when(priorityService.getCurrentPriorityBatch(operator.getOperatorId())).thenReturn(nodeEvents);
        EventOperatorDispatcher eventOperatorDispatcher =
                spy(new EventOperatorDispatcher(priorityService, null, operator, addonsManager,
                        operator.getNumberOfEventThreads(), ContextHelper.get()));
        when(priorityService.isDistributed()).thenReturn(true);
        doNothing().when(eventOperatorDispatcher).submitSweeper(any());
        ExecutorService executorService = Executors.newFixedThreadPool(1);

        executorService.submit(eventOperatorDispatcher);
        verify(eventOperatorDispatcher, timeout(TimeUnit.SECONDS.toMillis(10)).times(1)).submitSweeper(
                any());
    }

    public void testTraceIdAttachedByExecutor() throws ExecutionException, InterruptedException {
        when(operator.getOperatorId()).thenReturn("metadata-event-operator");
        when(operator.getNumberOfEventThreads()).thenReturn(1);
        when(operator.isOperatorEnabled()).thenReturn(true);

        EventOperatorDispatcher eventOperatorDispatcher =
                spy(new EventOperatorDispatcher(priorityService, null, operator, addonsManager,
                        operator.getNumberOfEventThreads(), ContextHelper.get()));

        ExecutorService executor = eventOperatorDispatcher.getExecutor();

        // The executor has access to the Tracer - the thread should create a new tracing span
        Future<?> invocation = executor.submit(() -> assertNotNull(artifactoryTracer.getCurrentThreadActiveSpan()));
        invocation.get();
    }

    public void testDifferentTraceIds() throws ExecutionException, InterruptedException {
        when(operator.getOperatorId()).thenReturn("metadata-event-operator");
        when(operator.getNumberOfEventThreads()).thenReturn(1);
        when(operator.isOperatorEnabled()).thenReturn(true);

        EventOperatorDispatcher eventOperatorDispatcher =
                spy(new EventOperatorDispatcher(priorityService, null, operator, addonsManager,
                        operator.getNumberOfEventThreads(), ContextHelper.get()));

        ExecutorService executor = eventOperatorDispatcher.getExecutor();

        Callable<String> c1 = () -> artifactoryTracer.getCurrentThreadActiveSpan().getSpan().context().toTraceId();
        Callable<String> c2 = () -> artifactoryTracer.getCurrentThreadActiveSpan().getSpan().context().toTraceId();
        List<Future<String>> futures = executor.invokeAll(List.of(c1, c2));

        assertNotEquals(futures.get(0).get(), futures.get(1).get());
    }
}