package org.artifactory.maven.index;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.index.IndexerDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.repo.RealRepo;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.mockito.ArgumentCaptor;
import org.springframework.util.ReflectionUtils;
import org.testng.Assert;
import org.testng.annotations.Test;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.assertTrue;

/**
 * @author Alexei Vainshtein
 */
public class MavenIndexerServiceImplTest extends ArtifactoryHomeBoundTest {

    @Test
    public void testCalcSpecificReposForIndexing() {
        MavenIndexerServiceImpl indexerService = new MavenIndexerServiceImpl();
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        indexerService.setRepositoryService(repositoryService);
        LocalRepoDescriptor local = new LocalRepoDescriptor();
        local.setKey("local");
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(List.of(local));
        FederatedRepoDescriptor fed = new FederatedRepoDescriptor();
        fed.setKey("test");
        when(repositoryService.getFederatedRepoDescriptors()).thenReturn(List.of(fed));
        Set<String> repos = indexerService.calcSpecificReposForIndexing(List.of("test"));
        Assert.assertEquals(repos.size(), 1);
    }

    @Test
    public void testSkipIndexingForNonExistingRepos() {
        MavenIndexerServiceImpl indexerService = spy(new MavenIndexerServiceImpl());
        // mock configDescriptor
        Field centralConfigField = ReflectionUtils.findField(MavenIndexerServiceImpl.class, "centralConfig");
        assertNotNull(centralConfigField);
        ReflectionUtils.makeAccessible(centralConfigField);
        CentralConfigService centralConfigService = mock(CentralConfigService.class);
        ReflectionUtils.setField(centralConfigField, indexerService, centralConfigService);
        CentralConfigDescriptorImpl centralConfigDescriptor = new CentralConfigDescriptorImpl();
        IndexerDescriptor mavenIndexer = new IndexerDescriptor();
        SortedSet<String> includedRepositories = new TreeSet<>();
        includedRepositories.add("does-not-exist-repo");
        mavenIndexer.setIncludedRepositories(includedRepositories);
        centralConfigDescriptor.setIndexer(mavenIndexer);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);

        // mock repositoryService
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        indexerService.setRepositoryService(repositoryService);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey("exists-repo");
        // TODO: 08/03/2022 replace with config service? layout comes from there.. 
        DbLocalRepo realRepo =
                new DbLocalRepo(localRepoDescriptor, localRepoDescriptor.toRepositoryConfig(), repositoryService,
                        null, null);
        ArrayList<RealRepo> realRepos = new ArrayList<>();
        realRepos.add(realRepo);
        when(repositoryService.getLocalFederatedAndRemoteRepositories()).thenReturn(realRepos);

        // when
        indexerService.index(new MavenIndexerRunSettings(true, true, null));

        // then
        ArgumentCaptor<List> listArgumentCaptor = ArgumentCaptor.forClass(List.class);
        verify(indexerService).mergeVirtualRepoIndexes(listArgumentCaptor.capture());
        assertTrue(listArgumentCaptor.getValue().isEmpty());
    }
}