package org.artifactory.metadata.service.indexer;

import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.metadata.service.MetadataEntityFacade;
import org.artifactory.metadata.service.MetadataEventService;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

class MetadataServiceIndexHandlerImplTest extends ArtifactoryHomeBoundJUnitTest {

    private MetadataServiceIndexHandlerImpl metadataServiceIndexHandler;
    @Mock
    private InternalArtifactoryContext context;
    @Mock
    private MetadataEntityFacade metadataEntityFacade;
    @Mock
    private MetadataEventService metadataEventService;
    @Mock
    private RepositoryService repositoryService;
    private AutoCloseable closeable;

    @BeforeEach
    void setUp() {
        closeable = MockitoAnnotations.openMocks(this);
        metadataServiceIndexHandler = Mockito.spy(new MetadataServiceIndexHandlerImpl());
        ArtifactoryContextThreadBinder.bind(context);
        Mockito.when(context.beanForType(InternalMetadataServiceIndexHandler.class))
                .thenReturn(metadataServiceIndexHandler);
        metadataServiceIndexHandler.setMetadataEventService(metadataEventService);
        metadataServiceIndexHandler.setRepositoryService(repositoryService);
        metadataServiceIndexHandler.setMetadataEntityFacade(metadataEntityFacade);
    }

    @AfterEach
    void tearDown() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        metadataServiceIndexHandler.destroy();
        closeable.close();
    }

    @Test
    void testIsPackageTypeSupportedWhenFeatureIsOff() {
        setRealTimeIndexingProp(false);
        Assertions.assertFalse(metadataServiceIndexHandler.isPackageTypeSupported("repo/some/path"));
    }

    @Test
    void testIsPackageTypeSupportedWithEmptyPath() {
        setRealTimeIndexingProp(true);
        Assertions.assertFalse(metadataServiceIndexHandler.isPackageTypeSupported(""));
    }

    @Test
    void testIsPackageTypeSupportedWithDocker() {
        setRealTimeIndexingProp(true);
        mockRepoAndPackageType("repo", RepoType.Docker);
        Assertions.assertTrue(metadataServiceIndexHandler.isPackageTypeSupported("repo/some/path"));
    }

    @Test
    void testIsPackageTypeSupportedWithGeneric() {
        setRealTimeIndexingProp(true);
        mockRepoAndPackageType("generic-local", RepoType.Generic);
        Assertions.assertFalse(metadataServiceIndexHandler.isPackageTypeSupported("generic-local"));
    }

    @Test
    void testShouldSkipIndexingPositive() {
        setRealTimeIndexingProp(true);
        mockRepoAndPackageType("generic-local", RepoType.Generic);
        Assertions.assertTrue(metadataServiceIndexHandler.shouldSkipIndexing(RepoPathFactory.create("generic-local")));

        RepoPath manifestPath = RepoPathFactory.create("docker-local/image/tag/manifest.json");
        Mockito.when(metadataEntityFacade.isLeadArtifact(manifestPath)).thenReturn(false);
        mockRepoAndPackageType("docker-local", RepoType.Docker);
        Assertions.assertTrue(metadataServiceIndexHandler.shouldSkipIndexing(manifestPath));
    }

    @Test
    void testShouldSkipIndexingNegative() {
        setRealTimeIndexingProp(true);
        RepoPath manifestPath = RepoPathFactory.create("docker-local/image/tag/manifest.json");
        Mockito.when(metadataEntityFacade.isLeadArtifact(manifestPath)).thenReturn(true);
        mockRepoAndPackageType("docker-local", RepoType.Docker);
        Assertions.assertFalse(metadataServiceIndexHandler.shouldSkipIndexing(manifestPath));
    }

    @Test
    void testDeletePackageAsync() {
        setRealTimeIndexingProp(true);
        mockRepoAndPackageType("docker-local", RepoType.Docker);
        RepoPath manifestPath = RepoPathFactory.create("docker-local/image/tag/manifest.json");
        Mockito.when(metadataEntityFacade.isLeadArtifact(manifestPath)).thenReturn(true);

        metadataServiceIndexHandler.deletePackageAsync(manifestPath);
        Mockito.verify(metadataServiceIndexHandler).submitPackageDeletionAfterCommit(manifestPath);
    }

    private void mockRepoAndPackageType(String repoKey, RepoType repoType) {
        RepoConfig repoConfig = Mockito.mock(RepoConfig.class);
        Mockito.when(repoConfig.getPackageType()).thenReturn(repoType);
        Mockito.when(repoConfig.getKey()).thenReturn(repoKey);
        Mockito.when(repositoryService.repoConfigByKey(repoKey)).thenReturn(repoConfig);
    }

    private void setRealTimeIndexingProp(boolean enabled) {
        getBound().setProperty(ConstantValues.metadataRealTimeIndexingEnable, String.valueOf(enabled));
        metadataServiceIndexHandler.init();
    }
}