package org.artifactory.metrics.providers.features;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.download.FolderDownloadConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedMember;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.trashcan.TrashcanConfigDescriptor;
import org.artifactory.repo.service.InternalRepositoryService;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.StreamSupport;

import static org.mockito.Mockito.when;

public class RepositoriesFeatureTest {

    @Mock
    private CentralConfigService centralConfigService;

    @Mock
    private InternalRepositoryService repositoryService;

    @Mock
    private CentralConfigDescriptor descriptor;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.openMocks(this);

        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        when(descriptor.getFederatedRepoUrlBaseOrDefault()).thenReturn("https://local");
        when(descriptor.getTrashcanConfig()).thenReturn(new TrashcanConfigDescriptor());
        when(descriptor.getFolderDownloadConfig()).thenReturn(new FolderDownloadConfigDescriptor());

        Map<String, FederatedRepoDescriptor> federatedRepoDescriptors = new HashMap<>();
        federatedRepoDescriptors.put("fed1",
                fedRepo("fed1", true, descriptor.getFederatedRepoUrlBaseOrDefault(), RepoType.Debian));
        federatedRepoDescriptors.put("fed2",
                fedRepo("fed2", false, descriptor.getFederatedRepoUrlBaseOrDefault(), RepoType.Docker));

        when(repositoryService.getFederatedRepoDescriptors()).thenReturn(
                new ArrayList<>(federatedRepoDescriptors.values()));
        List<RealRepoDescriptor> fedRepos = repositoryService.getFederatedRepoDescriptors().stream()
                .map(repo -> (RealRepoDescriptor) repo)
                .collect(Collectors.toList());
        when(repositoryService.getLocalFederatedAndRemoteRepoDescriptors()).thenReturn(fedRepos);
    }

    private FederatedRepoDescriptor fedRepo(String key, boolean enabled, String urlBase, RepoType type) {
        FederatedRepoDescriptor descriptor = new FederatedRepoDescriptor();
        descriptor.setKey(key);
        descriptor.setType(type);

        FederatedMember local = new FederatedMember();
        local.setUrl(urlBase + "/artifactory/" + key);
        local.setEnabled(enabled);

        FederatedMember remote1 = new FederatedMember();
        remote1.setUrl(urlBase + "/artifactory/" + key + "1");

        FederatedMember remote2 = new FederatedMember();
        remote2.setUrl("https://remote/artifactory/" + key + "2");

        descriptor.getFederatedMembers().add(local);
        descriptor.getFederatedMembers().add(remote1);
        descriptor.getFederatedMembers().add(remote2);

        return descriptor;
    }

    @Test
    public void testFederatedRepositoriesFeature() throws Exception {
        RepositoriesFeature repositoriesFeature = new RepositoriesFeature();
        FieldUtils.writeField(repositoriesFeature, "configService", centralConfigService, true);
        FieldUtils.writeField(repositoriesFeature, "repoService", repositoryService, true);
        JsonNode featureGroup = new ObjectMapper().valueToTree(repositoriesFeature.getFeature());

        Iterable<JsonNode> iterable = () -> featureGroup.get("features").elements();
        Optional<JsonNode> fedRepos = StreamSupport.stream(iterable.spliterator(), false)
                .filter(node -> node.get("name").asText().equals("federated repositories"))
                .findFirst();

        Assert.assertTrue(fedRepos.isPresent());
        Assert.assertEquals(fedRepos.get().get("attributes").get("number_of_repositories").asInt(), 2);

        Optional<JsonNode> fed1 = StreamSupport.stream(fedRepos.get().get("features").spliterator(), false)
                .filter(node -> node.get("name").asText().equals("fed1"))
                .findFirst();
        Assert.assertTrue(fed1.isPresent());
        Assert.assertEquals(fed1.get().get("attributes").get("enabled").asBoolean(), true);
        Assert.assertEquals(fed1.get().get("attributes").get("package_type").asText(), RepoType.Debian.name());
        Assert.assertEquals(fed1.get().get("attributes").get("number_of_mirrors").asInt(), 2);

        Optional<JsonNode> fed2 = StreamSupport.stream(fedRepos.get().get("features").spliterator(), false)
                .filter(node -> node.get("name").asText().equals("fed2"))
                .findFirst();
        Assert.assertTrue(fed2.isPresent());
        Assert.assertEquals(fed2.get().get("attributes").get("enabled").asBoolean(), false);
        Assert.assertEquals(fed2.get().get("attributes").get("package_type").asText(), RepoType.Docker.name());
        Assert.assertEquals(fed2.get().get("attributes").get("number_of_mirrors").asInt(), 2);

    }
}
