/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.metrics.services;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.Logger;
import ch.qos.logback.classic.spi.LoggingEvent;
import ch.qos.logback.core.Appender;
import org.apache.commons.lang3.reflect.FieldUtils;
import org.apache.http.impl.client.CloseableHttpClient;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.VersionInfo;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.metrics.callhome.CallHomeMetrics;
import org.artifactory.metrics.providers.features.CallHomeFeature;
import org.artifactory.security.access.AccessService;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.db.DbMetaData;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.jfrog.access.common.ServiceId;
import org.jfrog.storage.DbType;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.slf4j.LoggerFactory;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;

import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author Tomer Mayost
 */
public class CallHomeServiceTest {

    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private ArtifactoryApplicationContext artifactoryContext;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @Mock
    private CentralConfigDescriptor centralConfigDescriptor;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private CentralConfigService configService;
    @Mock
    private ArtifactoryServersCommonService serversService;
    @Mock
    private AccessService accessService;
    @Mock
    private DbService dbService;
    @Mock
    private AolService aolService;

    private CallHomeService callHomeService;
    private AutoCloseable closable;

    @BeforeMethod
    public void beforeMethod() {
        this.closable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ArtifactoryHome.bind(artifactoryHome);
        callHomeService =
                spy(new CallHomeService(addonsManager, configService, serversService, accessService, dbService,
                        aolService));
    }

    @AfterMethod
    public void afterMethod() throws Exception {
        ArtifactoryHome.unbind();
        ArtifactoryContextThreadBinder.unbind();
        this.closable.close();
    }

    @Test(dataProvider = "accountUrlExpectedAccountName")
    void checkCallHomeAccountName(String accountUrl, String expectedAccountName) {
        assertEquals(callHomeService.getAccountName(accountUrl), expectedAccountName);
    }

    @DataProvider
    private static Object[][] accountUrlExpectedAccountName() {
        return new Object[][] {
                {"https://bezirk.jfrog.io/bezirk/webapp/", "bezirk"},
                {"https://localhost:8080/artifactory/webapp/", "N/A"},
                {null, "N/A"}
        };
    }

    @Test(dataProvider = "testShouldExecuteCallHomeDataProvider")
    void testShouldExecuteCallHome(boolean isVersionQueryEnabled, boolean isOfflineMode, boolean isAol,
            boolean expectedShouldExecuteCallHome) {
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getBooleanProperty(ConstantValues.versionQueryEnabled))
                .thenReturn(isVersionQueryEnabled);
        when(configService.getDescriptor()).thenReturn(centralConfigDescriptor);
        when(centralConfigDescriptor.isOfflineMode()).thenReturn(isOfflineMode);
        when(aolService.isAol()).thenReturn(isAol);
        assertEquals(callHomeService.shouldExecuteCallHome(), expectedShouldExecuteCallHome);
    }

    @DataProvider
    private static Object[][] testShouldExecuteCallHomeDataProvider() {
        return new Object[][] {
                {true, true, true, true},
                {true, true, false, false},
                {true, false, true, true},
                {true, false, false, true},
                {false, true, true, false},
                {false, true, false, false},
                {false, false, true, false},
                {false, false, false, false}
        };
    }

    @Test
    public void testGetFeatureException() throws Exception {

        //Given
        Appender mockedAppender = Mockito.mock(Appender.class);
        ArgumentCaptor<LoggingEvent> loggingEventCaptor = ArgumentCaptor.forClass(LoggingEvent.class);
        Logger root = (Logger) LoggerFactory.getLogger(CallHomeService.class);
        root.addAppender(mockedAppender);
        root.setLevel(Level.ERROR);

        CentralConfigService centralConfigService = Mockito.mock(CentralConfigService.class);
        CentralConfigDescriptor descriptor = Mockito.mock(CentralConfigDescriptor.class);
        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(accessService.getArtifactoryServiceId())
                .thenReturn(new ServiceId(UUID.randomUUID().toString(), UUID.randomUUID().toString()));
        AolService aolService = Mockito.mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(artifactoryContext.beanForType(eq(AolService.class))).thenReturn(aolService);
        when(addonsManager.addonByType(eq(HaCommonAddon.class))).thenReturn(Mockito.mock(HaCommonAddon.class));
        when(dbService.getDatabaseType()).thenReturn(DbType.POSTGRESQL);
        DbMetaData dbMetadata = Mockito.mock(DbMetaData.class);
        when(dbMetadata.getProductVersion()).thenReturn("9.6");
        when(dbService.getDbMetaData()).thenReturn(dbMetadata);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(configService.getDescriptor()).thenReturn(centralConfigDescriptor);
        when(configService.getMutableDescriptor()).thenReturn(Mockito.mock(MutableCentralConfigDescriptor.class));
        VersionInfo versionInfo = new VersionInfo("7.x-SNAPSHOT", "2147483647");
        when(configService.getProductVersionInfo()).thenReturn(versionInfo);
        when(centralConfigDescriptor.isOfflineMode()).thenReturn(false);
        when(aolService.isAol()).thenReturn(true);

        CallHomeService service =
                new CallHomeService(addonsManager, configService, serversService, accessService, dbService,
                        aolService) {
                    @Override
                    CloseableHttpClient createHttpClient() {
                        return Mockito.mock(CloseableHttpClient.class);
                    }

                    @Override
                    boolean shouldExecuteCallHome() {
                        return true;
                    }
                };

        ((List<CallHomeFeature>) FieldUtils.readField(service, "callHomeFeatures", true))
                .add(() -> {
                    throw new RuntimeException("Get feature error !");
                });

        CallHomeMetrics callHomeMetrics = Mockito.mock(CallHomeMetrics.class);
        when(callHomeMetrics.getCallHomeData()).thenReturn(new ConcurrentHashMap<>());
        FieldUtils.writeField(service, "callHomeMetrics", callHomeMetrics, true);

        //When
        service.callHome();

        //Then
        verify(mockedAppender, times(1)).doAppend(loggingEventCaptor.capture());
        Assert.assertEquals(loggingEventCaptor.getValue().getLevel(), Level.ERROR);
        Assert.assertTrue(loggingEventCaptor.getValue().getMessage().startsWith("Failed to get feature from: "));
    }
}
