package org.artifactory.ph;

import com.jfrog.ph.acl.exception.PackageException;
import com.jfrog.ph.acl.exception.PackageForbiddenException;
import com.jfrog.ph.acl.exception.PackageNotFoundException;
import org.artifactory.api.repo.exception.BlackedOutException;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.repo.exception.SnapshotPolicyException;
import org.artifactory.fs.RepoResource;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.DownloadRequestContext;
import org.artifactory.request.RemoteRequestException;
import org.artifactory.resource.BlockedRepoResource;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.io.IOException;
import java.util.stream.Stream;

import static org.apache.http.HttpStatus.SC_CONFLICT;
import static org.apache.http.HttpStatus.SC_FORBIDDEN;
import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class ArtifactExtractorTest {

    private static final String REPO = "repo";
    private static final String PATH = "path";

    @Test
    void testLocateFileWithBlockedRepoResource() {
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        DownloadRequestContext context = mock(DownloadRequestContext.class);
        when(context.getRepoKey()).thenReturn(REPO);
        when(context.getPath()).thenReturn(PATH);
        Repo<?, ?> repo = mock(Repo.class);
        when(repositoryService.repositoryByKey(REPO)).thenReturn(repo);
        when(repo.getInfo(context)).thenReturn(new BlockedRepoResource(RepoPathFactory.create(REPO, PATH)));
        try {
            ArtifactExtractor.locateFile(context, repositoryService);
            Assertions.fail("Expected PackageForbiddenException");
        } catch (PackageForbiddenException e) {
            Assertions.assertEquals(SC_FORBIDDEN, e.getStatus());
            Assertions.assertEquals("Forbidden: artifact is blocked", e.getPublicReason());
        } catch (PackageException e) {
            Assertions.fail("Expected PackageForbiddenException");
        }
    }

    @ParameterizedTest
    @MethodSource("getResourceStreamHandleProvider")
    void getResourceStreamHandle(Exception catchException, PackageException expectedException)
            throws RepoRejectException, IOException {
        RepoResource info = mock(RepoResource.class);
        when(info.getRepoKey()).thenReturn(REPO);
        when(info.getPath()).thenReturn(PATH);
        InternalRepositoryService repositoryService = mock(InternalRepositoryService.class);
        DownloadRequestContext context = mock(DownloadRequestContext.class);
        Repo<?, ?> repo = mock(Repo.class);
        doThrow(catchException).when(repositoryService).getResourceStreamHandle(context, repo, info);
        PackageException exception = Assertions.assertThrows(PackageException.class,
                () -> ArtifactExtractor.getResourceStreamHandle(info, context, repo, repositoryService));
        Assertions.assertEquals(expectedException.getMessage(), exception.getMessage());
        Assertions.assertEquals(expectedException.getStatus(), exception.getStatus());
        Assertions.assertEquals(expectedException.getInternalReason(), exception.getInternalReason());
        Assertions.assertEquals(expectedException.getPublicReason(), exception.getPublicReason());
    }

    private static Stream<Arguments> getResourceStreamHandleProvider() {
        RepoConfig repoConfig = mock(RepoConfig.class);
        when(repoConfig.getKey()).thenReturn(REPO);
        RepoPath repoPath = RepoPathFactory.create(REPO, PATH);
        SnapshotPolicyException snapshotPolicyException = new SnapshotPolicyException(repoConfig, repoPath);
        String catchConflictExceptionMessage =
                String.format("Failed to find artifact in repo:%s, path:%s, cause:%s", REPO, PATH,
                        snapshotPolicyException.getMessage());
        BlackedOutException blackedOutException = new BlackedOutException(repoConfig, repoPath);
        String catchBlackedOutExceptionMessage =
                String.format("Failed to find artifact in repo:%s, path:%s, cause: repo is blacked out", REPO, PATH);
        RepoRejectException repoRejectException = new RepoRejectException("message");
        String catchRepoRejectExceptionMessage =
                String.format("Failed to find artifact in repo:%s, path:%s, cause:%s", REPO, PATH,
                        repoRejectException.getMessage());
        RemoteRequestException remoteRequestException = new RemoteRequestException("message", 402, "internal");
        String catchRemoteRequestExceptionMessage =
                String.format("Failed to find artifact in repo:%s, path:%s, cause:%s", REPO, PATH,
                        remoteRequestException.getMessage());
        IOException ioException = new IOException("stam");
        String catchIOExceptionMessage =
                String.format("Failed to find artifact, repo:%s, path:%s, cause:%s", REPO, PATH, ioException);
        return Stream.of(Arguments.of(snapshotPolicyException,
                        new PackageException(SC_CONFLICT, "Conflict", catchConflictExceptionMessage)),
                Arguments.of(blackedOutException, new PackageNotFoundException(catchBlackedOutExceptionMessage)),
                Arguments.of(repoRejectException, new PackageNotFoundException(catchRepoRejectExceptionMessage)),
                Arguments.of(remoteRequestException, new PackageNotFoundException(catchRemoteRequestExceptionMessage)),
                Arguments.of(ioException,
                        new PackageException(SC_INTERNAL_SERVER_ERROR, "Internal server error", catchIOExceptionMessage,
                                ioException)));
    }

}