/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo;

import org.apache.http.HttpEntity;
import org.apache.http.HttpStatus;
import org.apache.http.ProtocolVersion;
import org.apache.http.StatusLine;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpHead;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.message.BasicStatusLine;
import org.apache.http.protocol.HttpContext;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.addon.LayoutsCoreAddon;
import org.artifactory.addon.curation.CurationAddon;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.addon.plugin.PluginAction;
import org.artifactory.addon.plugin.PluginsAddon;
import org.artifactory.addon.plugin.download.BeforeRemoteDownloadAction;
import org.artifactory.addon.plugin.download.BeforeRemoteInfoRequestAction;
import org.artifactory.addon.smartrepo.EdgeSmartRepoAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.ResearchService;
import org.artifactory.api.repo.UpstreamRepoKind;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.mbean.MBeanRegistrationServiceImpl;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RemoteRepositoryRepoTypeConfig;
import org.artifactory.opentracing.ArtifactoryJaegerTracerConfiguration;
import org.artifactory.opentracing.ArtifactoryTracer;
import org.artifactory.repo.config.RepoConfigDefaultValues.DefaultUrl;
import org.artifactory.repo.http.DenyListCacheHolderService;
import org.artifactory.repo.http.GuaveCacheConnectionManagersHolder;
import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.repo.http.IdleConnectionMonitorService;
import org.artifactory.repo.http.IdleConnectionMonitorServiceImpl;
import org.artifactory.repo.http.mbean.HTTPConnectionPool;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.HeadersMultiMap;
import org.artifactory.request.Request;
import org.artifactory.request.RequestContext;
import org.artifactory.schedule.monitor.OnlineMonitorService;
import org.artifactory.storage.fs.service.FileService;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.artifactory.traffic.TrafficService;
import org.artifactory.util.RepoLayoutUtils;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.test.util.ReflectionTestUtils;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.List;

import static org.fest.assertions.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.atLeast;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.isNull;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Yoav Landman
 */
public class HttpRepoTest extends ArtifactoryHomeBoundJUnitTest {

    @Mock
    private InternalRepositoryService internalRepoService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private ResearchService researchService;
    @Mock
    private ArtifactoryContext contextMock;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private LayoutsCoreAddon layoutsCoreAddon;
    @Mock
    private HaAddon haAddon;
    @Mock
    private CurationAddon curationAddon;
    @Mock
    private HaCommonAddon haCommonAddon;
    @Mock
    private PluginsAddon pluginsAddon;
    @Mock
    private FileService fileService;
    @Mock
    private HttpConnectionPoolMonitorService httpConnectionPoolMonitorService;
    @Mock
    private MBeanRegistrationServiceImpl mBeanRegistrationService;
    @Mock
    private EdgeSmartRepoAddon edgeSmartRepoAddon;
    @Mock
    private AolService aolService;
    @Mock
    private TrafficService trafficService;

    private HttpRepoDescriptor descriptor;

    @BeforeEach
    public void setup() {
        System.setProperty(ConstantValues.artifactoryProductVersion.getPropertyName(),
                ArtifactoryServiceVersion.getCurrent().getVersion());
        MockitoAnnotations.initMocks(this);
        getBound().setProperty(ConstantValues.idleConnectionMonitorInterval, "10");
        getBound().setProperty(ConstantValues.disableIdleConnectionMonitoring, "false");
        IdleConnectionMonitorService idleConnectionMonitorService =
                new IdleConnectionMonitorServiceImpl(new GuaveCacheConnectionManagersHolder());
        when(addonsManager.addonByType(LayoutsCoreAddon.class)).thenReturn(layoutsCoreAddon);
        when(addonsManager.addonByType(HaAddon.class)).thenReturn(haAddon);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        when(addonsManager.addonByType(PluginsAddon.class)).thenReturn(pluginsAddon);
        when(addonsManager.addonByType(CurationAddon.class)).thenReturn(curationAddon);
        when(addonsManager.addonByType(EdgeSmartRepoAddon.class)).thenReturn(edgeSmartRepoAddon);
        when(contextMock.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(contextMock.beanForType(IdleConnectionMonitorService.class)).thenReturn(idleConnectionMonitorService);
        when(contextMock.beanForType(DenyListCacheHolderService.class)).thenReturn(
                mock(DenyListCacheHolderService.class));
        when(contextMock.beanForType(FileService.class)).thenReturn(fileService);
        when(contextMock.beanForType(MBeanRegistrationServiceImpl.class)).thenReturn(mBeanRegistrationService);
        when(contextMock.beanForType(HttpConnectionPoolMonitorService.class))
                .thenReturn(httpConnectionPoolMonitorService);
        when(contextMock.beanForType(TrafficService.class)).thenReturn(trafficService);
        when(aolService.isAol()).thenReturn(false);
        when(contextMock.beanForType(AolService.class)).thenReturn(aolService);
        when(researchService.isRepoConfiguredToSyncProperties(any(RemoteRepoConfig.class))).thenReturn(true);
        when(contextMock.beanForType(ArtifactoryTracer.class)).thenReturn(new ArtifactoryTracer(
                ArtifactoryJaegerTracerConfiguration.initTracer()));

        ArtifactoryContextThreadBinder.bind(contextMock);
        descriptor = spy(new HttpRepoDescriptor());
        when(edgeSmartRepoAddon.updateUpstreamRepoKind(any())).thenReturn(UpstreamRepoKind.SMART);
    }

    @AfterEach
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
        getBound().setProperty(ConstantValues.syncPropertiesBlacklistUrls, "");
    }

    @Test
    public void syncPropsDisabledWhenRepoConfigDisabled() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        httpRepo.initDefaultUrlsLists();
        when(researchService.isRepoConfiguredToSyncProperties(remoteRepoConfig)).thenReturn(false);
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();
    }

    @Test
    public void syncPropsEnabledForNonDefaultUrl() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Maven);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://13.11.201.9");//begin there
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        assertThat(httpRepo.isSynchronizeProperties()).isTrue();
    }

    @Test
    public void syncPropsEnabledForBlankDefaultUrl() {
        RemoteRepoConfig remoteRepoConfig1 = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig1.getRepoTypeConfig().setUrl("http://7.10.201.9");//end here
        HttpRepo httpRepo1 =
                new HttpRepo(descriptor, remoteRepoConfig1, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        assertThat(httpRepo1.isSynchronizeProperties()).isTrue();

        RemoteRepoConfig remoteRepoConfig2 = RemoteRepoConfig.create("repoKey", RepoType.GitLfs);
        remoteRepoConfig2.getRepoTypeConfig().setUrl("");

        HttpRepo httpRepo2 =
                new HttpRepo(descriptor, remoteRepoConfig1, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);

        assertThat(httpRepo2.isSynchronizeProperties()).isTrue();
    }

    @Test
    public void syncPropsDisabledForDefaultUrl() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        httpRepo.initDefaultUrlsLists();

        remoteRepoConfig.getRepoTypeConfig().setUrl(DefaultUrl.COCOAPODS.getUrl());
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        remoteRepoConfig.getRepoTypeConfig().setUrl(DefaultUrl.DEBIAN.getUrl());
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        remoteRepoConfig.getRepoTypeConfig().setUrl(DefaultUrl.BOWER.getUrl());
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        remoteRepoConfig.getRepoTypeConfig().setUrl(DefaultUrl.MAVEN.getUrl());
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        remoteRepoConfig.getRepoTypeConfig().setUrl(DefaultUrl.OPKG.getUrl());
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();
    }

    @Test
    public void syncPropertiesDisabledIfUrlInBlacklist() {
        String url = "http://19.8.201.9";//sign
        String anotherUrl = "http://5.10.201.4";//started here
        getBound().setProperty(ConstantValues.syncPropertiesBlacklistUrls, url);
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.BuildInfo);
        remoteRepoConfig.getRepoTypeConfig().setUrl(url);
        //Have to init repo again since const values are read on init method
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);

        httpRepo.initDefaultUrlsLists();
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        getBound().setProperty(ConstantValues.syncPropertiesBlacklistUrls, url + "," + anotherUrl);
        //Have to init repo again since const values are read on init method
        httpRepo.initDefaultUrlsLists();
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();

        remoteRepoConfig.getRepoTypeConfig().setUrl(anotherUrl);
        assertThat(httpRepo.isSynchronizeProperties()).isFalse();
    }

    @Test
    public void syncPropertiesEnabledIfUrlNotInBlacklist() {
        String url = "http://what.can.be.done.must.be/done";
        getBound().setProperty(ConstantValues.syncPropertiesBlacklistUrls, url);
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        assertThat(httpRepo.isSynchronizeProperties()).isTrue();
    }

    @Test
    public void testSuffixCreateHttpClient() throws IOException {
        when(centralConfigService.getDescriptor()).thenReturn(mock(CentralConfigDescriptor.class));
        RemoteRepoConfig repoConfig = RemoteRepoConfig.create("test-repo", RepoType.Generic);
        repoConfig.getRepoTypeConfig().setUrl("http://test");
        repoConfig.getRepoTypeConfig().setUsername("repo-username");
        repoConfig.getRepoTypeConfig().setPassword("repo-password");
        HttpRepo httpRepo =
                new HttpRepo(new HttpRepoDescriptor(), repoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);

        try (CloseableHttpClient ignored = httpRepo.createHttpClient(
                ConstantValues.httpClientMaxTotalConnections.getInt(), true, true, "test-suffix")) {
            verify(mBeanRegistrationService, times(1))
                    .register(any(HTTPConnectionPool.class), eq("HTTPConnectionPool"), eq("test-repo-test-suffix"));
        }
    }

    @Test
    public void testProxyRemoteAuthAndMultihome() throws IOException {
        String proxyKey = "proxy123";
        ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
        proxyDescriptor.setKey(proxyKey);
        proxyDescriptor.setHost("proxyHost");
        proxyDescriptor.setPort(12345);
        proxyDescriptor.setUsername("proxy-username");
        proxyDescriptor.setPassword("proxy-password");

        RemoteRepoConfig repoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        RemoteRepositoryRepoTypeConfig remoteRepoConfig = spy(new RemoteRepositoryRepoTypeConfig());
        ReflectionTestUtils.setField(repoConfig, "repoTypeConfig", remoteRepoConfig);
        repoConfig.getRepoTypeConfig().setUrl("http://test");
        repoConfig.getRepoTypeConfig().setProxyRef(proxyKey);
        repoConfig.getRepoTypeConfig().setUsername("repo-username");
        repoConfig.getRepoTypeConfig().setPassword("repo-password");
        repoConfig.getRepoTypeConfig().setLocalAddress("0.0.0.0");

        CentralConfigDescriptorImpl configDescriptor = spy(new CentralConfigDescriptorImpl());
        configDescriptor.setProxies(List.of(proxyDescriptor));
        when(this.centralConfigService.getDescriptor()).thenReturn(configDescriptor);
        HttpRepoDescriptor repoDescriptor = spy(new HttpRepoDescriptor());
        HttpRepo httpRepo =
                new HttpRepo(repoDescriptor, repoConfig, internalRepoService, addonsManager, researchService,
                        null, null, this.centralConfigService);
        CloseableHttpClient client =
                httpRepo.createHttpClient(ConstantValues.httpClientMaxTotalConnections.getInt(), true, true, null);
        if (client != null) {
            client.close();
        }
        verify(remoteRepoConfig, atLeast(1)).getUrl();
        verify(repoDescriptor, times(1)).getProxy();
        verify(remoteRepoConfig, times(2)).getUsername();
        verify(remoteRepoConfig, times(1)).getPassword();
        verify(remoteRepoConfig, times(1)).getLocalAddress();

        //TODO: [by YS] implement test on httpclient4
        /*Credentials proxyCredentials = client.getState().getProxyCredentials(AuthScope.ANY);
        Assert.assertNotNull(proxyCredentials);
        Assert.assertTrue(proxyCredentials instanceof UsernamePasswordCredentials,
                "proxyCredentials are of the wrong class");
        Assert.assertEquals(((UsernamePasswordCredentials) proxyCredentials).getUserName(), "proxy-username");
        Assert.assertEquals(proxyCredentials.getPassword(), "proxy-password");

        Credentials repoCredentials = client.getState().getCredentials(
                new AuthScope("test", AuthScope.ANY_PORT, AuthScope.ANY_REALM));
        Assert.assertNotNull(repoCredentials);
        Assert.assertTrue(repoCredentials instanceof UsernamePasswordCredentials,
                "repoCredentials are of the wrong class");
        Assert.assertEquals(((UsernamePasswordCredentials) repoCredentials).getUserName(), "repo-username");
        Assert.assertEquals(repoCredentials.getPassword(), "repo-password");

        Assert.assertEquals(client.getHostConfiguration().getLocalAddress().getHostAddress(), "0.0.0.0");*/
    }

    @Test
    public void shouldProcessWithoutExceptions() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://THIS_IS_A_DUMMY_URL/BLA/BLA/BLA");
        HttpRepo httpRepo =
                new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager, researchService, null,
                        null, centralConfigService);
        assertFalse(httpRepo.isNeedToNormalize());
    }

    @Test
    public void shouldNormalizeUrl() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://google.com");
        HttpRepo httpRepo =
                new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        assertTrue(httpRepo.isNeedToNormalize());
    }

    @Test
    public void shouldNotNormalizeUrl() {
        String[] hostsArray = ConstantValues.listOfReposHostsNeedToAvoidUrlNormalisation.getString().trim().split(",");
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        Arrays.stream(hostsArray).forEach(url -> {
            remoteRepoConfig.getRepoTypeConfig().setUrl("http://" + url.trim() + "/somepackage");
            HttpRepo httpRepo =
                    new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                            null, null, centralConfigService);
            assertFalse(httpRepo.isNeedToNormalize());
        });
    }

    @Test
    public void shouldHandleUrlNormalizationViaDescriptor() {
        // Check when disableUrlNormalization == true
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://someUrl/somePackage");
        remoteRepoConfig.getRepoTypeConfig().setDisableUrlNormalization(true);
        HttpRepo httpRepo =
                new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, centralConfigService);
        assertFalse(httpRepo.isNeedToNormalize());
        // Check when disableUrlNormalization == false
        remoteRepoConfig.getRepoTypeConfig().setDisableUrlNormalization(false);
        httpRepo = new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager,
                researchService, null, null, centralConfigService);
        assertTrue(httpRepo.isNeedToNormalize());
        // Check when wrong URL
        remoteRepoConfig.getRepoTypeConfig().setUrl("http:\\someUrl/somePackage");
        remoteRepoConfig.getRepoTypeConfig().setDisableUrlNormalization(true);
        httpRepo = new HttpRepo(null, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                null, null, centralConfigService);
        assertTrue(httpRepo.isNeedToNormalize());
    }

    @Test
    public void testPluginInvocationOnDownloadResource() throws IOException {
        // setup
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getBaseConfig().setRepoLayoutRef(RepoLayoutUtils.MAVEN_2_DEFAULT.getName());
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://localhost:8082/artifactory");
        when(layoutsCoreAddon.translateArtifactPath(any(), any(), any())).thenReturn("path/file.txt");
        when(fileService.exists(any())).thenReturn(true);
        CentralConfigDescriptorImpl configDescriptor = new CentralConfigDescriptorImpl();
        configDescriptor.setRepoLayouts(List.of(RepoLayoutUtils.MAVEN_2_DEFAULT));
        when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);

        HttpEntity httpEntity = mock(HttpEntity.class);
        when(httpEntity.getContent()).thenReturn(mock(InputStream.class));
        CloseableHttpResponse mockedResponse = mock(CloseableHttpResponse.class);
        when(mockedResponse.getStatusLine())
                .thenReturn(new BasicStatusLine(new ProtocolVersion("HTTP", 1, 1), HttpStatus.SC_OK, "OK"));
        when(mockedResponse.getEntity()).thenReturn(httpEntity);
        CloseableHttpClient client = mock(CloseableHttpClient.class);
        when(client.execute(any(HttpGet.class), (HttpContext) isNull())).thenReturn(mockedResponse);
        OnlineMonitorService onlineMonitorService = mock(OnlineMonitorService.class);
        HttpRepo httpRepo =
                spy(new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, onlineMonitorService, centralConfigService));
        when(onlineMonitorService.isResourceUnavailable(any(StatusLine.class))).thenReturn(false);
        doReturn(client).when(httpRepo).getHttpClient();
        httpRepo.init(false);

        ArgumentCaptor<Class<? extends PluginAction>> argCaptor = ArgumentCaptor.forClass(Class.class);

        // test
        httpRepo.downloadResource("path/file.txt");

        // verify
        verify(pluginsAddon, times(1)).execPluginActions(argCaptor.capture(), any(), any());

        List<Class<? extends PluginAction>> allClassesUsed = argCaptor.getAllValues();
        assertEquals(allClassesUsed.size(), 1);
        assertTrue(allClassesUsed.contains(BeforeRemoteDownloadAction.class));
    }

    @Test
    public void testPluginInvocationOnRetrieveInfo() throws IOException {
        // setup
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getBaseConfig().setRepoLayoutRef(RepoLayoutUtils.MAVEN_2_DEFAULT.getName());
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://localhost:8082/artifactory");
        when(layoutsCoreAddon.translateArtifactPath(any(), any(), any())).thenReturn("path/file.txt");
        when(fileService.exists(any())).thenReturn(true);
        CentralConfigDescriptorImpl configDescriptor = new CentralConfigDescriptorImpl();
        configDescriptor.setRepoLayouts(List.of(RepoLayoutUtils.MAVEN_2_DEFAULT));
        when(centralConfigService.getDescriptor()).thenReturn(configDescriptor);

        HttpEntity httpEntity = mock(HttpEntity.class);
        when(httpEntity.getContent()).thenReturn(mock(InputStream.class));
        CloseableHttpResponse mockedResponse = mock(CloseableHttpResponse.class);
        when(mockedResponse.getStatusLine())
                .thenReturn(new BasicStatusLine(new ProtocolVersion("HTTP", 1, 1), HttpStatus.SC_OK, "OK"));
        when(mockedResponse.getEntity()).thenReturn(httpEntity);
        CloseableHttpClient client = mock(CloseableHttpClient.class);
        when(client.execute(any(HttpHead.class), any(HttpContext.class))).thenReturn(mockedResponse);

        RequestContext requestContext = mock(RequestContext.class);
        Request request = mock(Request.class);
        when(requestContext.getRequest()).thenReturn(request);
        when(requestContext.alterPath(any(), any())).thenCallRealMethod();

        OnlineMonitorService onlineMonitorService = mock(OnlineMonitorService.class);
        HttpRepo httpRepo =
                spy(new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager,
                        researchService, null, onlineMonitorService, centralConfigService));
        doReturn(client).when(httpRepo).getHttpClient();
        httpRepo.init(false);

        ArgumentCaptor<Class<? extends PluginAction>> argCaptor = ArgumentCaptor.forClass(Class.class);

        // test
        httpRepo.retrieveInfo("path/file.txt", false, requestContext);

        // verify
        verify(pluginsAddon).execPluginActions(argCaptor.capture(), any(), any());
        verify(httpRepo).whiteListHeaders(requestContext);

        List<Class<? extends PluginAction>> allClassesUsed = argCaptor.getAllValues();
        assertEquals(allClassesUsed.size(), 1);
        assertTrue(allClassesUsed.contains(BeforeRemoteInfoRequestAction.class));
    }

    @Test
    public void testPlusSymbolEscape() {
        String actual = HttpRepo.encodeIfNeeded("http://localhost:8081/bellsoft-jre8u252+9-linux-amd64.tar.gz",
                null);
        assertEquals(actual, "http://localhost:8081/bellsoft-jre8u252%2B9-linux-amd64.tar.gz");
    }

    @Test
    public void testWhiteListHeadersWithNullContext() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        HttpRepo httpRepo =
                new HttpRepo(descriptor, remoteRepoConfig, internalRepoService, addonsManager, researchService,
                        null, null, null);
        HeadersMultiMap headersMultiMap = httpRepo.whiteListHeaders(null);
        assertNotNull(headersMultiMap);
    }

    @Test
    public void testCurationAddonInterceptorAdded() throws IOException {
        RemoteRepoConfig repoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        RemoteRepositoryRepoTypeConfig remoteRepoConfig = spy(new RemoteRepositoryRepoTypeConfig());
        ReflectionTestUtils.setField(repoConfig, "repoTypeConfig", remoteRepoConfig);
        repoConfig.getRepoTypeConfig().setUrl("http://test");
        repoConfig.getRepoTypeConfig().setUsername("repo-username");
        repoConfig.getRepoTypeConfig().setPassword("repo-password");
        repoConfig.getRepoTypeConfig().setLocalAddress("0.0.0.0");
        repoConfig.getRepoTypeConfig().setDisableProxy(true);
        HttpRepoDescriptor repoDescriptor = spy(new HttpRepoDescriptor());
        when(centralConfigService.getDescriptor()).thenReturn(mock(CentralConfigDescriptor.class));
        HttpRepo httpRepo =
                new HttpRepo(repoDescriptor, repoConfig, internalRepoService, addonsManager, researchService,
                        null, null, this.centralConfigService);
        when(curationAddon.isCurationWithTrustEnabled()).thenReturn(true);
        CloseableHttpClient client =
                httpRepo.createHttpClient(ConstantValues.httpClientMaxTotalConnections.getInt(), true, true, null);
        if (client != null) {
            client.close();
        }

        verify(curationAddon, times(1)).getCurationRequestInterceptor();
    }
}