package org.artifactory.util;

import org.apache.commons.io.input.CountingInputStream;
import org.apache.http.HttpStatus;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.ArtifactoryResponse;
import org.artifactory.api.request.InternalArtifactoryRequest;
import org.artifactory.api.request.ReplicationChecksumDeployRequest;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.md.Properties;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.ArtifactoryRequest;
import org.jfrog.storage.binstore.common.ReaderTrackingInputStream;
import org.jfrog.storage.binstore.ifc.ArtificialInputStream;
import org.jfrog.storage.binstore.ifc.UsageTracking;
import org.testng.annotations.Test;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;

import static org.artifactory.util.UploadServiceUtils.getInputStream;
import static org.artifactory.util.UploadServiceUtils.getTargetRepository;
import static org.artifactory.util.UploadServiceUtils.getTransferredSize;
import static org.artifactory.util.UploadServiceUtils.isSuccess;
import static org.artifactory.util.UploadServiceUtils.populateAndValidateItemPropertiesFromRequest;
import static org.artifactory.util.UploadServiceUtils.validateCanAnnotate;
import static org.artifactory.util.UploadServiceUtils.validateProperties;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;
import static org.testng.AssertJUnit.assertFalse;

/**
 * @author dudim
 */
public class UploadServiceUtilsTest {

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestChecksumDeployWithNoAnnotatePermission()
            throws RepoRejectException, UnsupportedEncodingException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        PropertiesImpl requestProps = new PropertiesImpl();
        requestProps.put("propKey", "propValue");
        when(requestMock.getProperties()).thenReturn(requestProps);
        when(requestMock.getHeader(ArtifactoryRequest.CHECKSUM_DEPLOY)).thenReturn("true");
        when(requestMock.getHeader(ArtifactoryRequest.CHECKSUM_SHA256))
                .thenReturn("15e470ec647ae0e6734ec2a397d8be444063aebab70fb6775c1a1e5f042b0ae6");
        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");
        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(false);
        UsageTracking usageTrackingMock = mock(UsageTracking.class);
        when(usageTrackingMock.incrementNoDeleteLock("SHA1")).thenReturn(1);
        ReaderTrackingInputStream readerTrackingInputStream =
                new ReaderTrackingInputStream(new StringInputStream("file body"), "SHA1", usageTrackingMock);
        Properties properties =
                populateAndValidateItemPropertiesFromRequest(requestMock, readerTrackingInputStream, repoPath,
                        authServiceMock);
        assertNotNull(properties);
        assertEquals(properties.size(), 1);
        assertEquals(properties.getFirst("sha256"), "15e470ec647ae0e6734ec2a397d8be444063aebab70fb6775c1a1e5f042b0ae6");
        verify(authServiceMock, times(1)).canAnnotate(repoPath);
    }

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestChecksumDeployValidScenario()
            throws RepoRejectException, UnsupportedEncodingException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        PropertiesImpl requestProps = new PropertiesImpl();
        requestProps.put("propKey", "propValue");
        when(requestMock.getProperties()).thenReturn(requestProps);
        when(requestMock.getHeader(ArtifactoryRequest.CHECKSUM_DEPLOY)).thenReturn("true");
        when(requestMock.getHeader(ArtifactoryRequest.CHECKSUM_SHA256))
                .thenReturn("15e470ec647ae0e6734ec2a397d8be444063aebab70fb6775c1a1e5f042b0ae6");
        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");
        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(true);
        UsageTracking usageTrackingMock = mock(UsageTracking.class);
        when(usageTrackingMock.incrementNoDeleteLock("SHA1")).thenReturn(1);
        ReaderTrackingInputStream readerTrackingInputStream =
                new ReaderTrackingInputStream(new StringInputStream("file body"), "SHA1", usageTrackingMock);
        Properties properties =
                populateAndValidateItemPropertiesFromRequest(requestMock, readerTrackingInputStream, repoPath,
                        authServiceMock);
        assertNotNull(properties);
        assertEquals(properties.size(), 2);
        assertEquals(properties.getFirst("sha256"), "15e470ec647ae0e6734ec2a397d8be444063aebab70fb6775c1a1e5f042b0ae6");
        assertEquals(properties.getFirst("propKey"), "propValue");
        verify(authServiceMock, times(1)).canAnnotate(repoPath);
    }

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestNoAnnotatePermission() throws RepoRejectException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        PropertiesImpl requestProps = new PropertiesImpl();
        requestProps.put("propKey", "propValue");
        when(requestMock.getProperties()).thenReturn(requestProps);

        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");

        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(false);
        Properties properties =
                populateAndValidateItemPropertiesFromRequest(requestMock, null, repoPath, authServiceMock);
        assertNull(properties);
        verify(authServiceMock, times(1)).canAnnotate(repoPath);
    }

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestWithNullProperties() throws RepoRejectException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        when(requestMock.getProperties()).thenReturn(null);

        AuthorizationService authServiceMock = mock(AuthorizationService.class);

        Properties properties = populateAndValidateItemPropertiesFromRequest(requestMock, null, null, authServiceMock);
        assertNull(properties);
        verify(authServiceMock, times(0)).canAnnotate(any());
    }

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestWithEmptyProperties() throws RepoRejectException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        when(requestMock.getProperties()).thenReturn(new PropertiesImpl());

        AuthorizationService authServiceMock = mock(AuthorizationService.class);

        Properties properties = populateAndValidateItemPropertiesFromRequest(requestMock, null, null, authServiceMock);
        assertNull(properties);
        verify(authServiceMock, times(0)).canAnnotate(any());
    }

    @Test(expectedExceptions = RepoRejectException.class)
    public void testPopulateAndValidateItemPropertiesFromRequestWithUnvalidProperties() throws RepoRejectException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        PropertiesImpl requestProps = new PropertiesImpl();
        requestProps.put("555propKey", "propValue");
        when(requestMock.getProperties()).thenReturn(requestProps);

        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");

        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(true);

        try {
            populateAndValidateItemPropertiesFromRequest(requestMock, null, repoPath, authServiceMock);
        } catch (RepoRejectException e) {
            assertEquals(e.getErrorCode(), HttpStatus.SC_BAD_REQUEST);
            throw e;
        }
    }

    @Test
    public void testPopulateAndValidateItemPropertiesFromRequestValidScenario() throws RepoRejectException {
        ArtifactoryRequest requestMock = mock(ArtifactoryRequest.class);
        PropertiesImpl requestProps = new PropertiesImpl();
        requestProps.put("propKey", "propValue");
        when(requestMock.getProperties()).thenReturn(requestProps);

        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");

        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(true);
        Properties properties = populateAndValidateItemPropertiesFromRequest(requestMock, null, repoPath,
                authServiceMock);

        assertNotNull(properties);
        assertEquals(properties.get("propKey").size(), 1);
        assertTrue(properties.get("propKey").contains("propValue"));
    }

    @Test
    public void testValidateCanAnnotateNoAnnotatePermission() throws RepoRejectException {
        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");
        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(false);
        assertFalse(validateCanAnnotate(repoPath, authServiceMock));
    }

    @Test
    public void testValidateCanAnnotateValidScenario() throws RepoRejectException {
        RepoPath repoPath = RepoPathFactory.create("repoKey", "path/to/uploaded/binary.tgz");
        AuthorizationService authServiceMock = mock(AuthorizationService.class);
        when(authServiceMock.canAnnotate(repoPath)).thenReturn(true);
        assertTrue(validateCanAnnotate(repoPath, authServiceMock));
    }

    @Test(expectedExceptions = RepoRejectException.class)
    public void testValidatePropertiesUnvalidProps() throws RepoRejectException {
        PropertiesImpl properties = new PropertiesImpl();
        properties.put("555propKey", "value");
        try {
            validateProperties(properties);
        } catch (RepoRejectException e) {
            assertEquals(e.getErrorCode(), HttpStatus.SC_BAD_REQUEST);
            throw e;
        }
    }

    @Test
    public void testValidatePropertiesValidProps() throws RepoRejectException {
        PropertiesImpl properties = new PropertiesImpl();
        properties.put("propKey", "value");
        validateProperties(properties);
    }

    @Test
    public void testGetTransferredSize() {
        ArtificialInputStream inputStream = new ArtificialInputStream("", "", "", 25L, "repo", null);
        assertEquals(getTransferredSize(inputStream), 25L);

        CountingInputStream countingInputStream = new CountingInputStream(inputStream);
        assertEquals(getTransferredSize(countingInputStream), 0L);
    }

    @Test
    public void testGetInputStream() throws IOException {
        ArtifactoryRequest request = mock(ArtifactoryRequest.class);
        ArtificialInputStream inputStream = new ArtificialInputStream("", "", "", 25L, "repo", null);
        when(request.getInputStream()).thenReturn(inputStream);
        assertEquals(getInputStream(request, null), inputStream);

        reset(request);
        InputStream dummy = mock(InputStream.class);
        when(request.getInputStream()).thenReturn(dummy);
        assertTrue(getInputStream(request, CountingInputStream::new) instanceof CountingInputStream);
    }

    @Test
    public void testShouldLogTrafficUponFailure() {
        assertFalse(isSuccess(null));

        ArtifactoryResponse response = mock(ArtifactoryResponse.class);
        when(response.getStatus()).thenReturn(400);
        assertFalse(isSuccess(response));

        reset(response);
        when(response.getStatus()).thenReturn(200);
        assertTrue(isSuccess(response));

        reset(response);
        when(response.getStatus()).thenReturn(201);
        assertTrue(isSuccess(response));
    }

    @Test
    public void testGetTargetRepositoryReplicationChecksumDeployRequest() {
        LocalRepo<?, ?> localRepo = mock(LocalRepo.class);
        when(localRepo.getKey()).thenReturn("cool-remote-cache");
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        when(internalRepositoryService.localFederatedOrCachedRepositoryByKey(any())).thenReturn(localRepo);
        RepoPath cacheRepoPath = InternalRepoPathFactory.create("cool-remote-cache", "cool.file");
        ReplicationChecksumDeployRequest replicationChecksumDeployRequest =
                new ReplicationChecksumDeployRequest(cacheRepoPath, true);
        LocalRepo<?, ?> targetRepository =
                getTargetRepository(replicationChecksumDeployRequest, internalRepositoryService);
        assertNotNull(targetRepository);
        assertEquals(targetRepository.getKey(), "cool-remote-cache");
        InternalArtifactoryRequest internalArtifactoryRequest = new InternalArtifactoryRequest(cacheRepoPath, true);
        targetRepository = getTargetRepository(internalArtifactoryRequest, internalRepositoryService);
        assertNull(targetRepository);
    }
}