package org.artifactory.util;

import lombok.experimental.UtilityClass;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.descriptor.repo.FederatedMember;
import org.jfrog.client.util.PathUtils;

import static org.artifactory.repo.federated.FederatedMember.BASE_URL_PLACEHOLDER;

/**
 * @author Tamar Rosen
 */
@UtilityClass
@Slf4j
public class FederatedMemberUtils {

    public static String resolveUrl(FederatedMember member) {
        return resolveUrl(member.getUrlOrTemplate());
    }

    public static String resolveUrl(org.artifactory.repo.federated.FederatedMember member) {
        return resolveUrl(member.getUrlOrTemplate());
    }

    private static String resolveUrl(String url) {
        if (!url.contains(BASE_URL_PLACEHOLDER)) {
            return url;
        }
        CentralConfigService configService = ContextHelper.get().getCentralConfig();
        if (configService == null || configService.getDescriptor() == null) {
            log.warn("Cannot resolve the local member's url '{}': Artifactory's config descriptor is not ready.", url);
            return url;
        }
        String baseUrl = configService.getDescriptor().getFederatedRepoUrlBaseOrDefault();
        if (baseUrl.endsWith("/artifactory")) {
            return StringUtils.replace(url, BASE_URL_PLACEHOLDER, baseUrl);
        } else {
            return StringUtils.replace(url, BASE_URL_PLACEHOLDER, baseUrl + "/artifactory");
        }
    }

    public static boolean isLocal(FederatedMember member, String key, String baseUrl) {
        String unresolvedUrl = member.getUrlOrTemplate();
        return isLocal(unresolvedUrl, key, baseUrl);
    }

    public static boolean isLocal(org.artifactory.repo.federated.FederatedMember member, String key, String baseUrl) {
        String unresolvedUrl = member.getUrlOrTemplate();
        return isLocal(unresolvedUrl, key, baseUrl);
    }

    private static boolean isLocal(String url, String key, String baseUrl) {
        String repoKey = PathUtils.getLastPathElement(url);
        return (url.contains(BASE_URL_PLACEHOLDER) || url.contains(baseUrl)) && key.equals(repoKey);
    }

    public static boolean isOldFormatLocal(org.artifactory.repo.federated.FederatedMember member, String repoKey,
            String baseUrl) {
        return isOldFormatLocal(member.getUrlOrTemplate(), repoKey, baseUrl);
    }

    public static boolean isOldFormatLocal(FederatedMember member, String repoKey, String baseUrl) {
        return isOldFormatLocal(member.getUrlOrTemplate(), repoKey, baseUrl);
    }

    private boolean isOldFormatLocal(String url, String key, String baseUrl) {
        if (org.apache.commons.lang.StringUtils.isBlank(baseUrl)) {
            throw new RuntimeException("Custom base URL must be configured for federated repositories' functionality");
        }
        String repoKey = PathUtils.getLastPathElement(url);
        return url.contains(baseUrl) && key.equals(repoKey);
    }

    public static void convertToNewLocalMember(org.artifactory.repo.federated.FederatedMember federatedMember,
            String baseUrl) {
        baseUrl = baseUrl.endsWith("/artifactory") ? baseUrl : baseUrl + "/artifactory";
        String convertedUrl = federatedMember.getUrlOrTemplate().replace(baseUrl, BASE_URL_PLACEHOLDER);
        federatedMember.setUrl(convertedUrl);
        log.debug("Converting local member: '{}' to contain a placeholder url: '{}'...",
                federatedMember.getUrlOrTemplate(), convertedUrl);
    }
}
