/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2019 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.logging.version;

import com.google.common.collect.Lists;
import org.apache.commons.io.FileUtils;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.environment.converter.local.version.v7.V700PreInitReplaceLogbackConverter;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.artifactory.version.ArtifactoryServiceVersionProvider;
import org.artifactory.version.XmlConverterUtils;
import org.artifactory.version.converter.XmlConverter;
import org.artifactory.version.v18.LogbackTrafficLogV2Converter;
import org.artifactory.version.v19.LogbackMetricLogNewAppenderConverter;
import org.artifactory.version.v20.LogbackMetricLogAppenderOrderConverter;
import org.artifactory.version.v21.LogbackCompressLatestConverter;
import org.artifactory.version.v215.LogbackMetricsLogsConverter;
import org.artifactory.version.v22.LogbackSizeAndIntervalPolicyConverter;
import org.artifactory.version.v23.LogbackRemoveStartArtifactoryProWithServicesLogger;
import org.artifactory.version.v24.LogbackOutboundRequestLogConverter;
import org.artifactory.version.v25.LogbackUsageConsumptionLogConverter;
import org.artifactory.version.v26.LogbackFeatureAdoptionLogConverter;
import org.artifactory.version.v27.LogbackFeatureAdoptionAppenderIntervalConverter;
import org.artifactory.version.v28.ConsoleLogsConverter;
import org.artifactory.version.v29.TargetBundleRetentionLogConverter;
import org.artifactory.version.v30.ConsoleLogsFilterConverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.List;

/**
 * Keeps track of the logging configuration versions
 * All converters before Artifactory 7 have been deleted, since we backup and replace the old backup file
 *
 * @author Noam Y. Tenne
 * @see V700PreInitReplaceLogbackConverter
 */
public enum LoggingVersion {
    v16(ArtifactoryServiceVersionProvider.v700m001.get()),
    v17(ArtifactoryServiceVersionProvider.v770m079.get(), new LogbackMetricsLogsConverter()),
    v18(ArtifactoryServiceVersionProvider.v777m006.get(), new LogbackTrafficLogV2Converter()),
    v19(ArtifactoryServiceVersionProvider.v7110m033.get(), new LogbackMetricLogNewAppenderConverter()),
    v20(ArtifactoryServiceVersionProvider.v7113m003.get(), new LogbackMetricLogAppenderOrderConverter()),
    v21(ArtifactoryServiceVersionProvider.v7150m001.get(), new LogbackCompressLatestConverter()),
    v22(ArtifactoryServiceVersionProvider.v7171m005.get(), new LogbackSizeAndIntervalPolicyConverter()),
    v23(ArtifactoryServiceVersionProvider.v7192m014.get(), new LogbackRemoveStartArtifactoryProWithServicesLogger()),
    v24(ArtifactoryServiceVersionProvider.v7200m020.get(), new LogbackOutboundRequestLogConverter()),
    v25(ArtifactoryServiceVersionProvider.v7250m001.get(), new LogbackUsageConsumptionLogConverter()),
    v26(ArtifactoryServiceVersionProvider.v731m007.get(), new LogbackFeatureAdoptionLogConverter()),
    v27(ArtifactoryServiceVersionProvider.v73710m001.get(), new LogbackFeatureAdoptionAppenderIntervalConverter()),
    v28(ArtifactoryServiceVersionProvider.v7420m001.get(), new ConsoleLogsConverter()),
    v29(ArtifactoryServiceVersionProvider.v7490m001.get(), new TargetBundleRetentionLogConverter()),
    v30(ArtifactoryServiceVersionProvider.v7520m001.get(), new ConsoleLogsFilterConverter());

    private static final Logger log = LoggerFactory.getLogger(LoggingVersion.class);

    private final ArtifactoryServiceVersion version;
    private final XmlConverter[] xmlConverters;

    /**
     * Main constructor
     *
     * @param from          Start version
     * @param xmlConverters XML converters required for the specified range
     */
    LoggingVersion(ArtifactoryServiceVersion from, XmlConverter... xmlConverters) {
        this.version = from;
        this.xmlConverters = xmlConverters;
    }

    XmlConverter[] getXmlConverters() {
        return xmlConverters;
    }

    /**
     * Run the needed conversions
     *
     * @param srcEtcDir the directory in which resides the logback file to convert
     */
    public static void convert(ArtifactoryServiceVersion version, File srcEtcDir, File targetEtcDir)
            throws IOException {
        // First create the list of converters to apply
        List<XmlConverter> converters = LoggingVersion.getEffectedXmlConverters(version);
        if (!converters.isEmpty()) {
            File logbackConfigFile = new File(srcEtcDir, ArtifactoryHome.LOGBACK_CONFIG_FILE_NAME);
            if (!logbackConfigFile.exists()) {
                log.info("Skipping logger converters as logback file doesn't exist.");
                return;
            }
            try {
                String result = XmlConverterUtils.convert(converters,
                        FileUtils.readFileToString(logbackConfigFile, StandardCharsets.UTF_8));
                backupAndSaveLogback(result, targetEtcDir);
            } catch (IOException e) {
                log.error("Error occurred while converting logback config for conversion: {}.", e.getMessage());
                log.debug("Error occurred while converting logback config for conversion", e);
                throw e;
            }
        }
    }

    /**
     * Creates a backup of the existing logback configuration file and proceeds to save post-conversion content
     *
     * @param result Conversion result
     * @param etcDir directory to which to save the conversion result
     */
    public static void backupAndSaveLogback(String result, File etcDir) throws IOException {
        File logbackConfigFile = new File(etcDir, ArtifactoryHome.LOGBACK_CONFIG_FILE_NAME);
        if (logbackConfigFile.exists()) {
            File originalBackup = new File(etcDir, "logback.original.xml");
            FileUtils.copyFile(logbackConfigFile, originalBackup);
        }
        FileUtils.writeStringToFile(logbackConfigFile, result, StandardCharsets.UTF_8);
    }

    public static void convert(ArtifactoryServiceVersion from, File path)
            throws IOException {
        boolean foundConversion = false;
        // All converters of versions above me needs to be executed in sequence
        LoggingVersion[] versions = LoggingVersion.values();
        for (LoggingVersion logingVersion : versions) {
            if (logingVersion.version.after(from)) {
                foundConversion = true;
                convert(from, path, path);
            }
        }
        // Write to log only if conversion has been executed
        if (foundConversion) {
            log.info("Ending database conversion from {}", from);
        }
    }

    public static List<XmlConverter> getEffectedXmlConverters(ArtifactoryServiceVersion importedVersion) {
        List<XmlConverter> converters = Lists.newArrayList();
        for (LoggingVersion loggingVersion : values()) {
            if (loggingVersion.version.afterOrEqual(importedVersion) && loggingVersion.xmlConverters != null) {
                converters.addAll(Lists.newArrayList(loggingVersion.xmlConverters));
            }
        }
        return converters;
    }
}
