package org.artifactory.storage.binstore.service;

import lombok.NonNull;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.checksum.ChecksumType;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.ReloadableBean;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.StorageException;
import org.jfrog.storage.authentication.AuthenticationTokenProvider;
import org.jfrog.storage.binstore.BinariesInfo;
import org.jfrog.storage.binstore.exceptions.BinaryNotFoundException;
import org.jfrog.storage.binstore.exceptions.BinaryRejectedException;
import org.jfrog.storage.binstore.exceptions.SignedUrlException;
import org.jfrog.storage.binstore.ifc.BinaryStoreFederatedMember;
import org.jfrog.storage.binstore.ifc.BinaryStream;
import org.jfrog.storage.binstore.ifc.GarbageCollectorInfo;
import org.jfrog.storage.binstore.ifc.ManualRedirectType;
import org.jfrog.storage.binstore.ifc.RedirectType;
import org.jfrog.storage.binstore.ifc.UsageTracking;
import org.jfrog.storage.binstore.ifc.model.BinaryElement;
import org.jfrog.storage.binstore.ifc.model.BinaryElementRequest;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersInfo;
import org.jfrog.storage.binstore.ifc.model.BinaryProvidersResponseMap;
import org.jfrog.storage.binstore.ifc.model.StorageInfo;
import org.jfrog.storage.binstore.manager.BinaryProviderManager;
import org.jfrog.storage.binstore.providers.base.BinariesTasksInfo;
import org.jfrog.storage.binstore.providers.base.PersistedRemoteTask;

import javax.annotation.Nullable;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * @author TamirH
 */
public interface ArtifactoryBinaryService extends ContextCreationListener, ReloadableBean, UsageTracking {
    /**
     * Retrieve the readable stream with the bytes associated with the provided SHA1 checksum.
     *
     * @param sha1    the checksum key
     * @param headers headers
     * @return the stream that should be closed by the user
     * @throws BinaryNotFoundException If the checksum does not exist in this store
     */
    @Deprecated
    InputStream getBinary(String sha1, Map<String, String> headers) throws BinaryNotFoundException;

    /**
     * Retrieve the readable stream with the bytes associated with the provided SHA1 checksum.
     *
     * @param element {@link BinaryElement} representing the required binary
     * @return the stream that should be closed by the user
     * @throws BinaryNotFoundException If the checksum does not exist in this store
     */
    InputStream getBinary(BinaryElement element) throws BinaryNotFoundException;

    /**
     * Retrieve the binary info without bytes associated with the provided SHA1 checksum for the purpose of
     * checksumDeploy.
     * If required by checkInFilestore flag it will try to create the missing binary in the DB before
     * returning the stream wrapper
     * @param element         {@link BinaryElement} representing the required binary info
     * @param checkInFilestore if true will try to create binary if not found in DB
     * @return a stream wrapper which contains required info without stream by itself
     * @throws BinaryNotFoundException  If the checksum does not exist
     * @throws StorageException         If an error occurred during this process
     */
    @NonNull
    InputStream getBinaryForChecksumDeploy(BinaryElement element, boolean checkInFilestore);

    RedirectType calcRedirectType(@NonNull RedirectType redirectType, boolean isAol,
            ManualRedirectType manualRedirectType);

    /**
     * Retrieves the download signed url from binary provider
     *
     * **Note**: Should be used only for Enterprise+ license.
     * Currently we don't check that inside, as its not the binary provider scope, and the current hack
     * is disgusting.
     * Please make sure to check that in the caller method.
     *
     * @throws SignedUrlException If couldn't generate signed url
     */
    @NonNull
    String getSignedUrl(@NonNull BinaryElement binaryElementRequest) throws SignedUrlException;

    /**
     * Retrieve all the information (sha1, sha2, MD5, length) for this SHA1 key.
     *
     * @param checksumType type of checksum to search by (md5 / sha1 / sha2)
     * @param checksum     the checksum key
     * @return The info if exists, null otherwise
     * @throws StorageException if an error happen accessing the binary store
     */
    @Nullable
    BinaryElement findBinary(ChecksumType checksumType, String checksum);

    /**
     * Retrieve a collection of binary information stored in this binary store,
     * based on a list of checksums keys.
     * The checksum type, SHA1 or MD5, will be determined by validation.
     * Any value that are invalid for all checksum type will be ignored.
     *
     * @param checksums list of checksum keys to look for
     * @return A distinct set of binaries information found, may be empty if nothing found
     * @throws StorageException if an error happen accessing the binary store
     */
    @NonNull
    Set<BinaryElement> findBinaries(@Nullable Collection<String> checksums);

    BinaryProvidersInfo<Map<String, String>> getBinaryProvidersInfo();

    /**
     * Add the whole content of the input stream in the binary store and close the input stream.
     * If the checksum for the whole stream already exists, the existing entry will be used.
     * The reference string is used for the reference count of this binary store.
     * Transitionally, the calling transaction will be suspended and all action on filesystem
     * will be done without DB TX opened. Then when binary data will need insert normal TX will be
     * done internally to this method call.
     * The input stream will not be consumed if the type is coming from here and the checksum is known.
     *
     * @param in the stream with all the bytes for the binary store
     * @return The info object for this binary entry
     * @throws java.io.IOException     if the bytes cannot be read from the stream or saved in binary store
     * @throws BinaryRejectedException If the Binary failed validation at the storage layer
     */
    @NonNull
    BinaryElement addBinary(BinaryStream in) throws IOException, BinaryRejectedException;

    /**
     * Check the sanity of this binary store.
     * DB, Filesystem or network access depending on the implementation.
     * Will throw exceptions if this binary store is not in full healthy state.
     *
     * @param localOnly when set to 'true' the ping will be done for local providers only and will not cascade to remote
     *                  providers if there are any.
     */
    void ping(boolean localOnly);

    BinariesInfo getBinariesInfo();

    /**
     * Return the total size managed by this binary store.
     *
     * @return The total size or -1 if the binary does not supports fetch of the full size.
     */
    long getStorageSize();

    /**
     * Force optimization once even if there is no need for optimization (For migration).
     */
    void forceOptimizationOnce();

    /**
     * Factory method for creating this interface's data model, {@link BinaryElement}
     */
    BinaryElement createBinaryElement(String sha1, String sha2, String md5, long length,
            Map<String, String> headers);

    /**
     * Factory method to create a {@link BinaryStream} which is used to insert binaries into the Binarystore.
     * Actual checksums and length are always calculated by the underlying store from the given stream {@param in},
     * The {@param info} may be provided as <b>expected</b> values which are used by the Binarystore for validation
     * (and failing insertion on errors if warranted).
     */
    BinaryStream createBinaryStream(InputStream in, Map<String, String> binaryHeaders);

    /**
     * Executing binary deletion
     *
     * @return true if the binary candidate was successfully deleted
     */
    boolean executeBinaryCleaner(GCCandidate candidate, GarbageCollectorInfo result);

    /**
     * @return true if instance has one of following types: Enterprise, Enterprise+ or Edge and deployed to AWS
     */
    boolean isCdnAllowed();

    void updateFederateRepoBinaryProvider(@NonNull Collection<BinaryStoreFederatedMember> membersToAdd,
            @NonNull Collection<BinaryStoreFederatedMember> membersToDelete,
            AuthenticationTokenProvider authenticationTokenProvider);

    /**
     * Add the binary data line to this data store based on all params after verification
     * from binary provider delegate.
     * Verification steps depends on configuration and delegates answers.
     * Method will returns null if no binary object with specified info is found,
     * or no filestore delegate was configured.
     *
     * @param sha1   The SHA1 key value
     * @param md5    The md5 of this new entry
     * @param length The length of the binary
     * @return The full binary info object created and stored, or null if no object
     * @throws BinaryNotFoundException if verification failed
     */
    @Nullable
    BinaryElement addBinaryRecord(Map<String, String> headers, String sha1, String sha2, String md5, long length)
            throws BinaryNotFoundException;

    /**
     * Activate Garbage Collection for this binary store
     *
     * @return The results of the GC process
     */
    GarbageCollectorInfo runFullGarbageCollect();


    void deleteBlobInfos(List<String> deletedBinaries);

    /**
     * Delete all left over folders and files in the binaries directory that are not declared
     * at all in this binary store.
     *
     * @param statusHolder A status holder of the process messages
     */
    void prune(BasicStatusHolder statusHolder);

    /**
     * The actual local directory where checksum binary files will be stored.
     *
     * @return the folder for this binary store
     * @deprecated Cannot be used anymore
     */
    @Nullable
    @Deprecated
    File getBinariesDir();

    /**
     * Indicates if the current binary provider is a cloud based and supports redirection
     */
    boolean isCloudProviderSupportingRedirectionConfigured();

    /**
     * Map each file provider to it's filestore folder
     *
     * @return A map of each filestore folder mapped to it's provider
     */
    StorageInfo getStorageInfoSummary();

    BinaryProviderManager getBinaryProviderManager();

    Collection<BinaryElement> findAllBinaries();

    /**
     * Deletes unreferenced archive paths. Shared archive paths might not be used after a binary is deleted.
     */
    int deleteUnusedArchivePaths();

    /**
     * Deletes unreferenced archive names. Shared archive names might not be used after a binary is deleted.
     */
    int deleteUnusedArchiveNames();

    /**
     * Increments the active users of a certain binary to prevent deletion while still in usage.
     *
     * @param sha1 The sha1 checksum to protect
     */
    @Override
    int incrementNoDeleteLock(String sha1);

    /**
     * Decrements the active users of a certain binary. Indicates that the active usage was ended.
     *
     * @param sha1 The sha1 checksum to remove protection from
     */
    @Override
    void decrementNoDeleteLock(String sha1);

    /**
     * BEWARE - this method is meant for a very specific usage - the sha2 db conversion.
     * NEVER EVER touch the binaries table outside of normal flow!
     */
    boolean updateSha2ForSha1(String targetSha1, String newSha2) throws SQLException;

    /**
     * Checks whether the requested {@param sha1} is in cloud provider - via cache of sha1 string only
     * (not the actual artifact), and verifies that redirect are allowed in that cloud provider.
     * Going out to real cloud provider only in case cache doesn't hold the sha1 in it.
     */
    boolean canCloudProviderGenerateRedirection(String sha1);

    /**
     * This method is meant for retention purposes only - restoring files from cold storage.
     *
     * Checks if file is restored from cold storage.
     * This will populate return object with a dedicated archive attribute to indicate the binary restore
     * state.
     *
     * @param request The binary request containing the restore candidate sha1
     * @return object with a dedicated archive attribute to indicate the binary restore state.
     */
    BinaryProvidersResponseMap isFileRestoredFromColdStorage(BinaryElementRequest request);

    /**
     * This method is meant for retention purposes only - restoring files from cold storage.
     *
     * Initiates the restore process in the scope of the archive binary provider.
     *
     * @param request The binary request containing the restore candidate sha1
     */
    void initiateRestoreFromColdStorage(BinaryElementRequest request);

    List<PersistedRemoteTask> getNextRemoteTaskFromDataDb(long batchSize, int maxExecutionsCount,
            long retryDelayMs,
            long maximumExecTimeMs) throws SQLException;

    boolean isCloudProviderConfigured();

    /**
     * @return binaries tasks status of the {@param repoKey} in the given {@param currentTime}
     */
    BinariesTasksInfo getBinariesTasksRepoStatus(String repoKey, long currentTime);

    /**
     * @return binaries tasks status
     */
    BinariesTasksInfo getBinariesTasksStatus();

    List<String> getAndManageErrors();
}
