package org.artifactory.storage.db.binstore.service.cleanup;

import com.google.common.collect.ImmutableList;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.storage.GCCandidate;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.storage.binstore.service.cleanup.GCFunction;
import org.artifactory.storage.binstore.service.cleanup.GCProvider;
import org.artifactory.storage.binstore.service.cleanup.GarbageCollectorStrategy;
import org.jfrog.storage.binstore.ifc.GarbageCollectorInfo;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.util.List;

import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

@Test
public class TrashAndBinariesGCProviderTest {

    private GCProvider provider;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private TrashService trashService;
    @Mock
    private ArtifactoryBinaryService binaryService;

    @BeforeClass
    private void setup() {
        MockitoAnnotations.initMocks(this);
        provider = new TrashAndBinariesGCProvider(repositoryService, trashService, binaryService);
    }

    private List<GCCandidate> getGcCandidates() {
        return ImmutableList.of(
                new GCCandidate(null, "aaa", "aaa", "aaa", 55),
                new GCCandidate(null, "bbb", "bbb", "bbb", 55),
                new GCCandidate(null, "ccc", "ccc", "ccc", 55),
                new GCCandidate(null, "ddd", "ddd", "ddd", 55)
        );
    }

    public void testGetAction() {
        GCFunction cleanupAction = provider.getAction();

        GarbageCollectorInfo gcInfo =
                new GarbageCollectorInfo(GarbageCollectorStrategy.TRASH_AND_BINARIES.name(), 10);
        List<GCCandidate> candidatesForDeletion = getGcCandidates();

        when(repositoryService.existsBySha1("aaa")).thenReturn(true);
        when(repositoryService.existsBySha1("bbb")).thenReturn(true);
        when(repositoryService.existsBySha1("ccc")).thenReturn(false);
        when(repositoryService.existsBySha1("ddd")).thenReturn(false);

        when(binaryService.executeBinaryCleaner(candidatesForDeletion.get(2), gcInfo))
                .thenReturn(true);
        when(binaryService.executeBinaryCleaner(candidatesForDeletion.get(3), gcInfo))
                .thenReturn(false);

        for (int i = 0; i < candidatesForDeletion.size(); i++) {
            GCCandidate candidate = candidatesForDeletion.get(i);
            boolean binaryDeleted = cleanupAction.accept(candidate, gcInfo);
            verify(trashService).undeployFromTrash(candidate);
            verify(repositoryService).existsBySha1(candidate.sha1());
            if (i < 2) {
                verify(binaryService, times(0))
                        .executeBinaryCleaner(candidate, gcInfo);
                Assert.assertFalse(binaryDeleted);
            } else {
                verify(binaryService, times(1))
                        .executeBinaryCleaner(candidate, gcInfo);
                if (i == 2) {
                    Assert.assertTrue(binaryDeleted);
                } else {
                    Assert.assertFalse(binaryDeleted);
                }
            }
        }
    }

    public void testGetBatch() {
        when(trashService.getGCCandidatesFromTrash()).thenReturn(getGcCandidates());
        List<GCCandidate> candidatesForDeletion = provider.getBatch();
        verify(trashService, times(1)).getGCCandidatesFromTrash();
        Assert.assertEquals(candidatesForDeletion.size(), 4);
    }

    public void testGetName() {
        Assert.assertEquals(GarbageCollectorStrategy.TRASH_AND_BINARIES.name(), provider.getName());
    }

    public void testShouldReportAfterBatch() {
        Assert.assertTrue(provider.shouldReportAfterBatch());
    }
}