package org.artifactory.storage.db.repo.service.mapper;

import com.google.gson.JsonParser;
import org.apache.commons.io.FileUtils;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.home.TestArtifactoryHome;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.jfrog.common.JsonUtils;
import org.testng.annotations.AfterClass;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.Test;

import java.io.File;
import java.io.IOException;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.math.BigInteger;
import java.nio.file.Files;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.UUID;

import static java.lang.reflect.Modifier.isFinal;
import static org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapperTest.RandomObjectPopulator.createAndFill;
import static org.testng.AssertJUnit.assertEquals;

/**
 * @author Tamar Rosen
 */
@Test
public class RepositoryConfigMapperTest {
    private RepositoryConfigMapperImpl mapper;

    @BeforeClass
    public void setup() throws Exception {
        createAndBindToArtifactoryHome();
        mapper = new RepositoryConfigMapperImpl();
    }

    @AfterClass
    public void afterClass() throws Exception {
        unbindAndCleanHome();
    }

    public void testLocalRepositoryConfigToDbRepositoryConfig() throws Exception {
        LocalRepoConfig originalRepoConfig = createAndFill(LocalRepoConfig.class);
        DbRepositoryConfig dbRepositoryConfig = mapper.repositoryConfigToDbRepositoryConfig(originalRepoConfig);
        RepoConfig afterMappingRepoConfig = mapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);

        String originalRepoConfigJson = JsonUtils.getInstance().valueToString(originalRepoConfig, true);
        String afterMappingRepoConfigJson = JsonUtils.getInstance().valueToString(afterMappingRepoConfig, true);

        assertEquals(JsonParser.parseString(originalRepoConfigJson),
                JsonParser.parseString(afterMappingRepoConfigJson));
    }

    public void testFederatedRepositoryConfigToDbRepositoryConfig() throws Exception {
        FederatedRepoConfig originalRepoConfig = createAndFill(FederatedRepoConfig.class);
        originalRepoConfig.getRepoTypeConfig().setCdnRedirect(true);
        originalRepoConfig.getRepoTypeConfig().setDownloadRedirect(true);

        DbRepositoryConfig dbRepositoryConfig = mapper.repositoryConfigToDbRepositoryConfig(originalRepoConfig);
        RepoConfig afterMappingRepoConfig = mapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);

        String originalRepoConfigJson = JsonUtils.getInstance().valueToString(originalRepoConfig, true);
        String afterMappingRepoConfigJson = JsonUtils.getInstance().valueToString(afterMappingRepoConfig, true);

        assertEquals(JsonParser.parseString(originalRepoConfigJson),
                JsonParser.parseString(afterMappingRepoConfigJson));
    }

    public void testRemoteRepositoryConfigToDbRepositoryConfig() throws Exception {
        RemoteRepoConfig originalRepoConfig = createAndFill(RemoteRepoConfig.class);
        DbRepositoryConfig dbRepositoryConfig = mapper.repositoryConfigToDbRepositoryConfig(originalRepoConfig);
        RepoConfig afterMappingRepoConfig = mapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);

        String originalRepoConfigJson = JsonUtils.getInstance().valueToString(originalRepoConfig, true);
        String afterMappingRepoConfigJson = JsonUtils.getInstance().valueToString(afterMappingRepoConfig, true);

        assertEquals(JsonParser.parseString(originalRepoConfigJson),
                JsonParser.parseString(afterMappingRepoConfigJson));
    }

    public void testVirtualRepositoryConfigToDbRepositoryConfig() throws Exception {
        VirtualRepoConfig originalRepoConfig = createAndFill(VirtualRepoConfig.class);
        DbRepositoryConfig dbRepositoryConfig = mapper.repositoryConfigToDbRepositoryConfig(originalRepoConfig);
        RepoConfig afterMappingRepoConfig = mapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);

        String originalRepoConfigJson = JsonUtils.getInstance().valueToString(originalRepoConfig, true);
        String afterMappingRepoConfigJson = JsonUtils.getInstance().valueToString(afterMappingRepoConfig, true);

        assertEquals(JsonParser.parseString(originalRepoConfigJson),
                JsonParser.parseString(afterMappingRepoConfigJson));
    }

    public void testDbRepositoryConfigToRepositoryConfig() {
        DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("my-repo", RepoType.Conan);
        dbRepositoryConfig.setKey("my-repo");
        dbRepositoryConfig.setType("local");
        dbRepositoryConfig.setPackageType("conan");
        repositoryConfig.getBaseConfig().setModelVersion(1);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        byte[] bytes = JsonUtils.getInstance().valueToByteArray(repositoryConfig);
        dbRepositoryConfig.setRepositoryInfoBlob(bytes);

        RepoConfig repositoryConfig1 = mapper.dbInfoBlobToRepositoryConfig(dbRepositoryConfig);
        DbRepositoryConfig dbRepositoryConfig1 = mapper.repositoryConfigToDbRepositoryConfig(repositoryConfig1);

        String dbRepoConfig1Json = JsonUtils.getInstance().valueToString(dbRepositoryConfig1, true);
        String dbRepoConfigJson = JsonUtils.getInstance().valueToString(dbRepositoryConfig, true);

        assertEquals(JsonParser.parseString(dbRepoConfigJson), JsonParser.parseString(dbRepoConfig1Json));
    }

    private void createAndBindToArtifactoryHome() throws IOException {
        File homeDir = Files.createTempDirectory(this.getClass().getSimpleName()
                + "_" + new Random().nextLong()).toFile();
        ArtifactoryHome artifactoryHome = new TestArtifactoryHome(homeDir);
        artifactoryHome.initArtifactorySystemProperties();
        ArtifactoryHome.bind(artifactoryHome);
    }

    private void unbindAndCleanHome() {
        if (ArtifactoryHome.isBound()) {
            FileUtils.deleteQuietly(ArtifactoryHome.get().getHomeDir());
            ArtifactoryHome.unbind();
        }
    }

    public static class RandomObjectPopulator {
        private static final Random random = new Random();

        public static <T> T createAndFill(Class<T> clazz) throws Exception {
            T instance;
            if (RepoConfig.class.isAssignableFrom(clazz)) {
                Constructor<T> declaredConstructor = clazz.getDeclaredConstructor(String.class, RepoType.class);
                declaredConstructor.setAccessible(true);
                instance = declaredConstructor
                        .newInstance(getRandomValueForType(String.class), getRandomValueForType(RepoType.class));
            } else {
                instance = clazz.newInstance();
            }
            for (Field field : clazz.getDeclaredFields()) {
                if (isFinal(field.getModifiers())) {
                    continue;
                }
                field.setAccessible(true);
                Object value = getRandomValueForType(field.getType());
                field.set(instance, value);
            }
            return instance;
        }

        private static Object getRandomValueForType(Class<?> type) throws Exception {
            if (type.isEnum()) {
                Object[] enumValues = type.getEnumConstants();
                return enumValues[random.nextInt(enumValues.length)];
            } else if (type.equals(Integer.TYPE) || type.equals(Integer.class)) {
                return random.nextInt();
            } else if (type.equals(Long.TYPE) || type.equals(Long.class)) {
                return random.nextLong();
            } else if (type.equals(Double.TYPE) || type.equals(Double.class)) {
                return random.nextDouble();
            } else if (type.equals(Float.TYPE) || type.equals(Float.class)) {
                return random.nextFloat();
            } else if (type.equals(String.class)) {
                return UUID.randomUUID().toString();
            } else if (type.equals(BigInteger.class)) {
                return BigInteger.valueOf(random.nextInt());
            } else if (type.equals(Boolean.TYPE) || type.equals(Boolean.class)) {
                return random.nextBoolean();
            } else if (type.equals(List.class)) {
                return List.of();
            } else if (type.equals(Map.class)) {
                return Map.of();
            }
            return createAndFill(type);
        }
    }
}