package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.fetch;

import com.google.common.collect.Lists;
import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.api.rest.common.model.continues.FetchFunction;
import org.artifactory.api.rest.common.model.continues.util.PagingUtils;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.projects.ProjectsService;
import org.artifactory.projects.ProjectsUtils;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.TreeFilter;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.artifactory.util.CollectionUtils;
import org.jfrog.common.StreamSupportUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static org.artifactory.descriptor.repo.RepoType.ReleaseBundles;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE_REPO;
import static org.artifactory.repo.RepoPath.REMOTE_CACHE_SUFFIX;
import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.VirtualRemoteRepositoryNode.fromRemoteDescriptor;
import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.VirtualRemoteRepositoryNode.fromVirtualDescriptor;
import static org.artifactory.util.bundles.BundleConfigurationUtil.DEFAULT_RELEASE_BUNDLES_V2_REPOSITORY;

/**
 * @author Omri Ziv
 */
public abstract class RootFetchStrategy {

    private static final Logger log = LoggerFactory.getLogger(RootFetchStrategy.class);
    TreeFilter treeFilter;
    RepositoryService repositoryService;
    String projectKey;
    private String nextContinueState;
    private final boolean defaultContext;
    private final SystemRepoFetchService fetchService;
    final DescriptorFilter descriptorFilter;

    RootFetchStrategy(TreeFilter treeFilter) {
        this.treeFilter = treeFilter;
        ArtifactoryContext artifactoryContext = ContextHelper.get();
        this.repositoryService = artifactoryContext.getRepositoryService();
        this.projectKey = Optional.ofNullable(treeFilter.getProjectKey())
                .filter(StringUtils::isNotEmpty)
                .orElse(ProjectsService.DEFAULT_PROJECT);
        defaultContext = ProjectsUtils.isDefaultProject(projectKey);
        this.descriptorFilter = new DescriptorFilter(treeFilter);
        this.fetchService = new SystemRepoFetchService(treeFilter, defaultContext, descriptorFilter);
    }

    static <T extends RepoDescriptor> int getLastIndexOfKeys(List<String> originalRepoKeys,
            List<T> filteredRepoDescriptor,
            int limit) {
        return CollectionUtils.notNullOrEmpty(filteredRepoDescriptor) && filteredRepoDescriptor.size() == limit
                ? originalRepoKeys.indexOf(filteredRepoDescriptor.get(filteredRepoDescriptor.size() - 1).getKey()) + 1
                : originalRepoKeys.size();
    }

    static <T extends RepoDescriptor> List<T> sortAndGet(List<T> original) {
        return StreamSupportUtils.stream(original)
                .sorted(new RootFetchStrategy.RepoComparator())
                .collect(Collectors.toList());
    }

    public ContinueResult<RepositoryNode> fetchItems() {
        log.debug("Fetching repository items");
        TreeFilter nextTreeFilter = new TreeFilter(treeFilter);
        if (StringUtils.isNotBlank(nextContinueState)) {
            nextTreeFilter.setContinueIndex(Long.parseLong(nextContinueState));
        }
        ContinueResult<RepositoryNode> result =
                PagingUtils.getPagingFromMultipleFunctions(nextTreeFilter, getFetchFunctions());
        sortUpReleaseBundlesV2(result);
        log.debug("{} repositories fetched. continue state was set to: '{}'",
                result.getData().size(), result.getContinueState());
        nextContinueState = result.getContinueState();
        return result;
    }

    private void sortUpReleaseBundlesV2(@NonNull ContinueResult<RepositoryNode> result) {
        List<RepositoryNode> releaseBundlesV2 =
                result.getData().stream().filter(this::isReleaseBundleV2).toList();
        //keep the default project release-bundles-v2 repository always as a top repository.
        List<RepositoryNode> releaseBundlesV2Sorted = releaseBundlesV2.stream().sorted((item1, item2) -> {
            if (DEFAULT_RELEASE_BUNDLES_V2_REPOSITORY.equals(item1.getRepoKey())) {return -1;}
            if (DEFAULT_RELEASE_BUNDLES_V2_REPOSITORY.equals(item2.getRepoKey())) {return 1;}
            return 0;
        }).toList();
        result.getData().removeAll(releaseBundlesV2Sorted);
        //release-bundles-v2 repositories coming after release-bundles-v1
        int targetPosition = 0;
        for (int index = 0; index < result.getData().size(); index++) {
            if (isReleaseBundleV1(result.getData().get(index))) {
                targetPosition = index + 1;
            } else if (targetPosition != 0) {
                break;
            }
        }
        result.getData().addAll(targetPosition, releaseBundlesV2Sorted);
    }

    private boolean isReleaseBundleV1(@NonNull RepositoryNode repositoryNode) {
        return RELEASE_BUNDLE_REPO.equals(repositoryNode.getRepoType());
    }

    private boolean isReleaseBundleV2(@NonNull RepositoryNode repositoryNode) {
        return !isReleaseBundleV1(repositoryNode) && ReleaseBundles == repositoryNode.getRepoPkgType();
    }

    protected List<FetchFunction<RepositoryNode>> getFetchFunctions() {
        return new ArrayList<>(Arrays.asList(new FetchFunction<>(fetchService::getReleaseBundlesRepoNodes,
                        (long) repositoryService.getReleaseBundlesRepoDescriptors().size()),
                new FetchFunction<>(fetchService::getTrashRepo),
                new FetchFunction<>(fetchService::getSupportBundles)));
    }

    List<RepositoryNode> getNodes(List<RepoDescriptor> repos) {
        List<RepositoryNode> items = Lists.newArrayListWithCapacity(repos.size());
        repos.forEach(repo -> {
            String repoType = repo.getKey().endsWith(REMOTE_CACHE_SUFFIX) ? "cached"
                    : RepositoryType.byDescriptor(repo).getTypeNameLoweCase();
            if (repo instanceof VirtualRepoDescriptor) {
                items.add(fromVirtualDescriptor((VirtualRepoDescriptor) repo));
            } else if (repo instanceof RemoteRepoDescriptor) {
                items.add(fromRemoteDescriptor((RemoteRepoDescriptor) repo));
            } else {
                items.add(new RepositoryNode(repo.getKey(), repo.getType(), repoType));
            }
        });
        return items;
    }

    protected boolean isDefaultProjectContext() {
        return defaultContext;
    }

    static class RepoComparator implements Comparator<RepoDescriptor> {

        @Override
        public int compare(RepoDescriptor descriptor1, RepoDescriptor descriptor2) {

            //Local repositories can be either ordinary or caches
            if (descriptor1 instanceof LocalRepoDescriptor) {
                boolean repo1IsCache = ((LocalRepoDescriptor) descriptor1).isCache();
                boolean repo2IsCache = ((LocalRepoDescriptor) descriptor2).isCache();

                //Cache repositories should appear in a higher priority
                if (repo1IsCache && !repo2IsCache) {
                    return 1;
                } else if (!repo1IsCache && repo2IsCache) {
                    return -1;
                }
            }
            return descriptor1.getKey().compareTo(descriptor2.getKey());
        }
    }
}
