package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.fetch;

import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.api.rest.common.model.continues.FetchFunction;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.rest.common.model.RestModel;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.TreeFilter;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.VirtualRemoteRepositoryNode;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * @author Omri Ziv
 */
public class RootFetchByPackageTypeTest extends RootFetchStrategyTest {

    @Test
    public void testByType() {
        createMocks();
        TreeFilter treeFilter = new TreeFilter();
        treeFilter.setPackageTypes(Arrays.asList(RepoType.Maven, RepoType.Docker));
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();
        assertEquals(fetchFunctions.size(), 5);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 5L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 10L);

        ContinueResult<? extends RestModel> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 5);
        ContinueResult<? extends RestModel> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 10, 10);
    }

    @Test(dataProvider = "provideProjectKeys")
    public void testByRepoType(String project) {
        createMocks();
        TreeFilter treeFilter = new TreeFilter();
        treeFilter.setProjectKey(project);
        treeFilter.setPackageTypes(List.of(RepoType.Npm));
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();
        assertEquals(fetchFunctions.size(), 4);

        ContinueResult<? extends RestModel> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 9);
        assertTrue(continueResult1.getData().get(7) instanceof VirtualRemoteRepositoryNode);
        assertTrue(continueResult1.getData().get(8) instanceof VirtualRemoteRepositoryNode);
        long virtualRemotes = continueResult1.getData().stream()
                .filter(repo -> repo instanceof VirtualRemoteRepositoryNode).count();
        assertEquals(virtualRemotes, 2);
    }

    @Test
    public void testByTypeNoFilter() {
        createMocks();

        TreeFilter treeFilter = new TreeFilter();
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 7);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 5L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 10L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 7L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 5);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 10);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 9);

    }

    @Test
    public void testByTypeIncludingLocalAndVirtual() {
        createMocksWithLocalAndVirtual();

        TreeFilter treeFilter = new TreeFilter();
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 9);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 13);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 8);

    }

    @Test
    public void testByTypeIncludingLocalAndVirtualCheckSkipAndLimit() {
        createMocksWithLocalAndVirtual();

        TreeFilter treeFilter = new TreeFilter();
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(2, 1);
        assertEquals(continueResult1.getData().size(), 1);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 3);
        assertEquals(continueResult2.getData().size(), 3);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 4);
        assertEquals(continueResult3.getData().size(), 4);
        ContinueResult<RepositoryNode> continueResult4 = fetchFunctions.get(5).getFunction().apply(3, 51);
        assertEquals(continueResult4.getData().size(), 5);

    }

    @Test
    public void testByTypeOnlyVirtualFilter() {
        createMocksWithLocalAndVirtual();

        TreeFilter treeFilter = new TreeFilter();
        treeFilter.setRepositoryTypes(Collections.singletonList(RepositoryType.VIRTUAL));
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 4);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 3);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 1);
    }

    @Test
    public void testByTypeOnlyRemoteExpectNone() {
        createMocksWithLocalAndVirtual();

        TreeFilter treeFilter = new TreeFilter();
        treeFilter.setRepositoryTypes(Collections.singletonList(RepositoryType.REMOTE));
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 0);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 0);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 0);
    }

    @Test
    public void testByTypeFilterByRepoKey() {
        createMocksWithLocalAndVirtual();

        TreeFilter treeFilter = new TreeFilter();
        treeFilter.setByRepoKey("00");
        RootFetchByPackageType rootFetchByPackageType = new RootFetchByPackageType(treeFilter);
        List<FetchFunction<RepositoryNode>> fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        ContinueResult<RepositoryNode> continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 9);
        ContinueResult<RepositoryNode> continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 12);
        ContinueResult<RepositoryNode> continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 8);

        treeFilter.setByRepoKey("001");
        fetchFunctions = rootFetchByPackageType.getFetchFunctions();

        assertEquals(fetchFunctions.size(), 6);
        assertEquals(fetchFunctions.get(3).getOriginalSize().longValue(), 9L);
        assertEquals(fetchFunctions.get(4).getOriginalSize().longValue(), 13L);
        assertEquals(fetchFunctions.get(5).getOriginalSize().longValue(), 8L);

        continueResult1 = fetchFunctions.get(3).getFunction().apply(0, 20);
        assertEquals(continueResult1.getData().size(), 2);
        continueResult2 = fetchFunctions.get(4).getFunction().apply(0, 20);
        assertEquals(continueResult2.getData().size(), 2);
        continueResult3 = fetchFunctions.get(5).getFunction().apply(0, 20);
        assertEquals(continueResult3.getData().size(), 2);
    }
}
