package org.artifactory.ui.rest.service.admin.configuration.layouts;

import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.ui.rest.model.admin.configuration.layouts.RepositoryLayoutModel;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author Alexei Vainshtein
 */
public class GetLayoutInfoServiceTest {

    private static final String testLayoutName = "default";
    private static final List<String> federatedReposWithTestedLayout = List.of("test4", "test5");
    private static final List<String> localReposWithTestedLayout = List.of("test", "test2");
    private static final String fedRepoWithoutTestLayout = "test6";
    private static final String localRepoWithoutTestLayout = "test3";

    @Test
    public void testCreateAssociations() {
        RepoLayout defaultLayout = createRepoLayout(testLayoutName);
        RepositoryService repositoryService = mock(RepositoryService.class);
        createLocalRepositories(defaultLayout, repositoryService);
        createFederatedRepositories(defaultLayout, repositoryService);

        GetLayoutInfoService service = new GetLayoutInfoService(null, repositoryService);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.isAdmin()).thenReturn(true);
        RepositoryLayoutModel layoutModel = new RepositoryLayoutModel(defaultLayout);
        service.setAuthorizationService(authorizationService);
        service.createAssociations(layoutModel);
        assertEquals(layoutModel.getRepositoryAssociations().getLocalRepositories().size(), 4);
        verify(authorizationService, times(3)).isAdmin();
        verifyNoMoreInteractions(authorizationService);
    }

    @Test
    public void testCreateAssociationsWithLimitedUser() {
        RepoLayout defaultLayout = createRepoLayout(testLayoutName);
        RepositoryService repositoryService = mock(RepositoryService.class);
        createLocalRepositories(defaultLayout, repositoryService);
        createFederatedRepositories(defaultLayout, repositoryService);

        GetLayoutInfoService service = new GetLayoutInfoService(null, repositoryService);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(authorizationService.isAdmin()).thenReturn(false);
        when(authorizationService.isAnyProjectAdmin()).thenReturn(true);
        when(authorizationService.isProjectAdmin(new RepoPathImpl(localReposWithTestedLayout.get(0), "")))
                .thenReturn(true);
        when(authorizationService.isProjectAdmin(new RepoPathImpl(federatedReposWithTestedLayout.get(0), "")))
                .thenReturn(true);

        RepositoryLayoutModel layoutModel = new RepositoryLayoutModel(defaultLayout);
        service.setAuthorizationService(authorizationService);
        service.createAssociations(layoutModel);
        assertEquals(layoutModel.getRepositoryAssociations().getLocalRepositories().size(), 2);
        verify(authorizationService, times(3)).isAdmin();
        verify(authorizationService, times(4)).isProjectAdmin(any(RepoPathImpl.class));
        verifyNoMoreInteractions(authorizationService);
    }

    private void createFederatedRepositories(RepoLayout defaultLayout, RepositoryService repositoryService) {
        List<FederatedRepoDescriptor> federatedRepoDescriptors = new ArrayList<>();
        federatedRepoDescriptors
                .add(createFederatedRepo(federatedReposWithTestedLayout.get(0), RepoType.Generic, defaultLayout));
        federatedRepoDescriptors
                .add(createFederatedRepo(federatedReposWithTestedLayout.get(1), RepoType.Npm, defaultLayout));
        federatedRepoDescriptors
                .add(createFederatedRepo(fedRepoWithoutTestLayout, RepoType.Docker, createRepoLayout("non-def")));
        when(repositoryService.getFederatedRepoDescriptors()).thenReturn(federatedRepoDescriptors);
    }

    private void createLocalRepositories(RepoLayout defaultLayout, RepositoryService repositoryService) {
        List<LocalRepoDescriptor> localRepoDescriptors = new ArrayList<>();
        localRepoDescriptors.add(createLocalRepo(localReposWithTestedLayout.get(0), RepoType.Generic, defaultLayout));
        localRepoDescriptors.add(createLocalRepo(localReposWithTestedLayout.get(1), RepoType.Docker, defaultLayout));
        localRepoDescriptors
                .add(createLocalRepo(localRepoWithoutTestLayout, RepoType.Npm, createRepoLayout("non-def")));
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(localRepoDescriptors);
    }

    private LocalRepoDescriptor createLocalRepo(String key, RepoType type, RepoLayout repoLayout) {
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey(key);
        localRepoDescriptor.setType(type);
        localRepoDescriptor.setRepoLayout(repoLayout);
        return localRepoDescriptor;
    }

    private FederatedRepoDescriptor createFederatedRepo(String key, RepoType type, RepoLayout repoLayout) {
        FederatedRepoDescriptor federatedRepoDescriptor = new FederatedRepoDescriptor();
        federatedRepoDescriptor.setKey(key);
        federatedRepoDescriptor.setType(type);
        federatedRepoDescriptor.setRepoLayout(repoLayout);
        return federatedRepoDescriptor;
    }

    private RepoLayout createRepoLayout(String name) {
        RepoLayout repoLayout = new RepoLayout();
        repoLayout.setName(name);
        return repoLayout;
    }
}