package org.artifactory.ui.rest.service.admin.configuration.repositories.util;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.ph.PackageSpecification;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.LocalRepositoryInfo;
import org.artifactory.ui.rest.model.admin.configuration.repository.info.RepositoryInfo;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectResourceStatus;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.anyBoolean;
import static org.mockito.Mockito.anySet;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * @author BarakH
 */
@Test
public class GetRepositoryInfoServiceTest {

    private enum UserType {SIMPLE, PROJECT_ADMIN, ADMIN;}

    private static final String REPO_TYPE = "repoType";
    private static final String PROJECT_A = "proj1";
    private static final String PROJECT_B = "proj2";
    private static final String REPO_A = "repo-a";
    private static final String DEFAULT_PROJECT_REPO = "DEFAULT_PROJECT_REPO";

    @Mock
    private RepositoryService repositoryService;
    @Mock
    private CentralConfigService configService;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private PackageSpecification packageSpecification;

    GetRepositoryInfoService getRepositoryInfoService;

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        getRepositoryInfoService =
                new GetRepositoryInfoService(repositoryService, configService, projectsService, authorizationService,
                        packageSpecification);
        when(packageSpecification.hasReindex(any(), any())).thenReturn(true);
        LocalRepoDescriptor localRepo = new LocalRepoDescriptor();
        localRepo.setKey("A");
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(List.of(localRepo));
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        when(descriptor.getMultiLocalReplications("A")).thenReturn(List.of());
        when(configService.getDescriptor()).thenReturn(descriptor);
    }

    @Test(dataProvider = "provideUserType")
    public void testGetRepositoriesInfo(UserType userType, String project, boolean shouldBeExist) {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey(REPO_TYPE)).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(project);
        switch (userType) {
            case SIMPLE:
                when(authorizationService.isAdmin()).thenReturn(false);
                when(authorizationService.isProjectAdmin(project)).thenReturn(false);
                break;
            case PROJECT_ADMIN:
                when(authorizationService.isAdmin()).thenReturn(false);
                when(authorizationService.isProjectAdmin(project)).thenReturn(true);
                when(projectsService.getProjectRepositoriesKeys(PROJECT_A, true)).thenReturn(Set.of("A", "B"));
                break;
            case ADMIN:
                when(authorizationService.isAdmin()).thenReturn(true);
                when(authorizationService.isProjectAdmin(project)).thenReturn(true);
                when(projectsService.getProjectRepositoriesKeys(PROJECT_A, true)).thenReturn(Set.of("A", "B"));
                break;
        }
        getRepositoryInfoService.execute(restRequest, response);
        List<RepositoryInfo> result = (List<RepositoryInfo>) response.getiModelList();
        if (shouldBeExist) {
            assertEquals("A", result.get(0).getRepoKey());
        } else {
            assertTrue(result.isEmpty());
        }
    }

    @Test
    public void testExecuteRepoInfoIsPopulatedWithProjectInfo() {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey(REPO_TYPE)).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(PROJECT_A);
        when(authorizationService.isProjectAdmin(PROJECT_A)).thenReturn(true);
        when(projectsService.getProjectRepositoriesKeys(PROJECT_A, true)).thenReturn(Set.of(REPO_A));

        LocalRepoDescriptor repoA = new LocalRepoDescriptor();
        repoA.setType(RepoType.Maven);
        repoA.setKey(REPO_A);
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(List.of(repoA));
        when(projectsService.getRepoKeyToProjectMap(Set.of(REPO_A), true)).thenReturn(Map.of(REPO_A, PROJECT_A));
        when(projectsService.getProjectRepositoriesMap(PROJECT_A, true))
                .thenReturn(Map.of(REPO_A, ProjectResourceEntity.builder().resourceName(PROJECT_A)
                        .resourceStatus(ProjectResourceStatus.SHARED).sharedWithProject(PROJECT_B).build()));
        getRepositoryInfoService.execute(restRequest, response);

        assertEquals(response.getiModelList().size(), 1);
        LocalRepositoryInfo repoInfo = (LocalRepositoryInfo) response.getiModelList().stream().findFirst().orElse(null);
        assertEquals(repoInfo.getProjectKey(), PROJECT_A);
        assertNotNull(repoInfo.getEnvironments());
        assertEquals(repoInfo.getSharedWithProjects(), Set.of(PROJECT_B));
        assertFalse(repoInfo.isShareWithAllProjects());

        verify(projectsService).getRepoKeyToProjectMap(Set.of(REPO_A), true);
        verify(projectsService).getProjectRepositoriesKeys(PROJECT_A, true);
        verify(projectsService).getProjectRepositoriesMap(PROJECT_A, true);
        verify(projectsService).getProjectDisplayName(PROJECT_A);
        verifyNoMoreInteractions(projectsService);
    }

    @Test
    public void testExecuteRepoInfoIsPopulatedDefaultProjectInfo() {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey(REPO_TYPE)).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(null);
        when(authorizationService.isAdmin()).thenReturn(true);
        mockProjectsData();

        getRepositoryInfoService.execute(restRequest, response);
        assertEquals(response.getiModelList().size(), 2);
        List<LocalRepositoryInfo> repoInfo =
                (List<LocalRepositoryInfo>) response.getiModelList().stream().collect(Collectors.toList());

        LocalRepositoryInfo repoA = repoInfo.get(0);
        assertEquals(repoA.getProjectKey(), ProjectsService.DEFAULT_PROJECT);
        assertEquals(repoA.getProjectName(), ProjectsService.DEFAULT_PROJECT);
        assertNotNull(repoA.getEnvironments());
        assertEquals(repoA.getSharedWithProjects(), Set.of(PROJECT_A));
        assertTrue(repoA.isShareWithAllProjects());
        assertTrue(repoA.isSharedReadOnly());
        assertTrue(repoA.isTarget());

        LocalRepositoryInfo repoB = repoInfo.get(1);
        assertEquals(repoB.getProjectKey(), PROJECT_A);
        assertEquals(repoB.getProjectName(), PROJECT_A);
        assertNotNull(repoB.getEnvironments());
        assertEquals(repoB.getSharedWithProjects(), Set.of());
        assertFalse(repoB.isShareWithAllProjects());
        assertFalse(repoB.isSharedReadOnly());
        assertFalse(repoB.isTarget());

        verify(projectsService).getRepoKeyToProjectMap(Set.of(REPO_A, DEFAULT_PROJECT_REPO), true);
        verify(projectsService).getAllProjectsRepositoriesMap(true);
        verify(projectsService).getProjectDisplayName(PROJECT_A);
        verify(projectsService).getProjectDisplayName(ProjectsService.DEFAULT_PROJECT);
        verifyNoMoreInteractions(projectsService);
    }

    @Test
    public void testExecuteRepoInfoIsPopulatedWithUnassignedProject() {
        ArtifactoryRestRequest restRequest = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(restRequest.getPathParamByKey(REPO_TYPE)).thenReturn(RepoDetailsType.LOCAL_REPO);
        when(restRequest.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(ProjectsService.UNASSIGNED);
        when(repositoryService.getAllRepoKeys()).thenReturn(Set.of(REPO_A, DEFAULT_PROJECT_REPO));
        when(authorizationService.isAdmin()).thenReturn(true);
        mockProjectsData();

        getRepositoryInfoService.execute(restRequest, response);
        assertEquals(response.getiModelList().size(), 1);
        List<LocalRepositoryInfo> repoInfo =
                (List<LocalRepositoryInfo>) response.getiModelList().stream().collect(Collectors.toList());

        LocalRepositoryInfo repoA = repoInfo.get(0);
        assertEquals(repoA.getProjectKey(), ProjectsService.DEFAULT_PROJECT);
        assertEquals(repoA.getProjectName(), ProjectsService.DEFAULT_PROJECT);
        assertNotNull(repoA.getEnvironments());
        assertEquals(repoA.getSharedWithProjects(), Set.of(PROJECT_A));
        assertTrue(repoA.isShareWithAllProjects());
        assertTrue(repoA.isSharedReadOnly());

        verify(projectsService, times(2)).getRepoKeyToProjectMap(anySet(), anyBoolean());
        verify(projectsService).getAllProjectsRepositoriesMap(true);
        verify(projectsService).getProjectDisplayName(ProjectsService.DEFAULT_PROJECT);
        verifyNoMoreInteractions(projectsService);
    }

    private void mockProjectsData() {
        when(projectsService.getProjectRepositoriesKeys(ProjectsService.DEFAULT_PROJECT, true))
                .thenReturn(Set.of(DEFAULT_PROJECT_REPO));
        LocalRepoDescriptor repo = new LocalRepoDescriptor();
        repo.setType(RepoType.Maven);
        repo.setKey(DEFAULT_PROJECT_REPO);
        LocalRepoDescriptor repo2 = new LocalRepoDescriptor();
        repo2.setType(RepoType.Maven);
        repo2.setKey(REPO_A);
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(List.of(repo, repo2));
        when(projectsService.getRepoKeyToProjectMap(Set.of(REPO_A, DEFAULT_PROJECT_REPO), true))
                .thenReturn(Map.of(REPO_A, PROJECT_A, DEFAULT_PROJECT_REPO, ProjectsService.DEFAULT_PROJECT));
        when(projectsService.getAllProjectsRepositoriesMap(true)).thenReturn(Map.of(ProjectsService.DEFAULT_PROJECT,
                Map.of(DEFAULT_PROJECT_REPO,
                        ProjectResourceEntity.builder()
                                .resourceName(DEFAULT_PROJECT_REPO)
                                .sharedWithProject(PROJECT_A)
                                .resourceStatus(ProjectResourceStatus.SHARED_WITH_ALL_TARGET)
                                .sharedReadOnly(true)
                                .build()),
                PROJECT_A, Map.of(REPO_A, ProjectResourceEntity.builder()
                        .resourceName(PROJECT_A)
                        .resourceStatus(ProjectResourceStatus.SHARED)
                        .sharedReadOnly(false)
                        .build())));
        when(projectsService.getProjectDisplayName(PROJECT_A)).thenReturn(PROJECT_A);
        when(projectsService.getProjectDisplayName(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(ProjectsService.DEFAULT_PROJECT);
    }

    @DataProvider
    public static Object[][] provideUserType() {
        return new Object[][] {
                {UserType.SIMPLE, "", false},
                {UserType.SIMPLE, PROJECT_A, false},
                {UserType.PROJECT_ADMIN, "", false},
                {UserType.PROJECT_ADMIN, PROJECT_A, true},
                {UserType.ADMIN, "", true},
                {UserType.ADMIN, PROJECT_A, true},
        };
    }
}