package org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder;

import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.engine.DownloadRedirectHelper;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.ProxyRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.local.LocalRepoChecksumPolicyType;
import org.artifactory.repo.packagetype.SnapshotVersionBehavior;
import org.artifactory.storage.binstore.service.ArtifactoryBinaryService;
import org.artifactory.ui.rest.model.admin.configuration.repository.BasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.CdnRedirectRepoConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.RepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepoAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederationConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteAdvancedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteBasicRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteCacheRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteNetworkRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.DockerTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.NpmTypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.typespecific.TypeSpecificConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualBasicRepositoryConfigModel;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.validation.constraints.NotNull;
import java.util.List;

import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.config.RepoConfigDefaultValues.DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.mockito.Answers.RETURNS_DEEP_STUBS;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * Author: Ivan Perepelytsya
 * Date: 6/3/20.
 */
@Test
public class RepoConfigBuilderTest {

    @Mock
    private DownloadRedirectHelper downloadRedirectHelper;
    @Mock
    private ArtifactoryBinaryService binaryService;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private MutableCentralConfigDescriptor mutableCentralConfigDescriptor;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @Mock
    private RepositoryConfigService repositoryConfigService;
    @InjectMocks
    private RepoConfigDescriptorBuilder repoConfigBuilder;
    private TypeSpecificConfigModel typeSpecificConfigModel;

    private AutoCloseable mockitoAutoCloseable;

    @BeforeMethod
    public void initMocks() {
        mockitoAutoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(mutableCentralConfigDescriptor.getCurrentReverseProxy()).thenReturn(null);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(centralConfigService.getMutableDescriptor()).thenReturn(mutableCentralConfigDescriptor);
        when(downloadRedirectHelper.isRedirectEnableForThisRepoType(anyString())).thenReturn(true);
        when(binaryService.isCdnAllowed()).thenReturn(true);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
    }

    @AfterMethod
    private void cleanup() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
        mockitoAutoCloseable.close();
    }

    @Test
    public void testBuildLocalDescriptorWithCdnRedirectEnabled() {
        typeSpecificConfigModel = new NpmTypeSpecificConfigModel();
        final LocalRepositoryConfigModel model = prepareLocalModel();
        LocalConfig localConfig = repoConfigBuilder.buildLocalConfig(model);
        assertNotNull(localConfig.getRepoTypeConfig().getCdnRedirectRepoConfig());
        assertTrue(localConfig.getRepoTypeConfig().getCdnRedirectRepoConfig().isEnabled());
    }

    @Test
    public void testPopulateProxy() {
        String proxyKey = "some-proxy-key";
        CentralConfigDescriptor descriptor = mock(CentralConfigDescriptor.class);
        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        List.of(prepareFederatedRepoModel(), prepareRemoteModel()).forEach(model -> {
            ProxyRepoConfig repoConfig = buildRepoConfig(model);
            assertNotNull(repoConfig);
            //proxy does not exist
            assertNull(repoConfig.getProxyRef());
            // create default proxy
            ProxyDescriptor proxyDescriptor = new ProxyDescriptor();
            proxyDescriptor.setKey(proxyKey);
            proxyDescriptor.setPlatformDefault(true);
            when(descriptor.getProxy(proxyKey)).thenReturn(proxyDescriptor);

            setProxyKey(model, proxyKey);
            setDisableProxy(model, true);
            repoConfig = buildRepoConfig(model);
            assertNotNull(repoConfig);

            // proxy is disabled
            assertNull(repoConfig.getProxyRef());

            setDisableProxy(model, false);
            // default proxy should not be set as proxy ref
            assertNull(repoConfig.getProxyRef());

            proxyDescriptor.setPlatformDefault(false);
            repoConfig = buildRepoConfig(model);
            assertNotNull(repoConfig);
            // a non default proxy should be set as proxy ref
            assertNotNull(repoConfig.getProxyRef());
            assertEquals(repoConfig.getProxyRef(), proxyKey);
        });
    }

    private ProxyRepoConfig buildRepoConfig(RepositoryConfigModel model) {
        if (model instanceof RemoteRepositoryConfigModel remoteRepositoryConfigModel) {
            return repoConfigBuilder.buildRemoteRepoConfig(remoteRepositoryConfigModel);
        }
        if (model instanceof FederatedRepositoryConfigModel federatedRepositoryConfigModel) {
            return repoConfigBuilder.buildFederatedRepoConfig(federatedRepositoryConfigModel);
        }
        Assert.fail(String.format("failed to repo config from model '%s'", model.getClass().getSimpleName()));
        return null;
    }

    private void setProxyKey(RepositoryConfigModel model, String proxyKey) {
        if (model instanceof RemoteRepositoryConfigModel remoteRepositoryConfigModel) {
            remoteRepositoryConfigModel.getAdvanced().getNetwork().setProxy(proxyKey);
        } else if (model instanceof FederatedRepositoryConfigModel federatedRepositoryConfigModel) {
            federatedRepositoryConfigModel.getAdvanced().setProxy(proxyKey);
        } else {
            Assert.fail(String.format("failed to set proxy key to model '%s'", model.getClass().getSimpleName()));
        }
    }

    private void setDisableProxy(RepositoryConfigModel model, boolean disableProxy) {
        if (model instanceof RemoteRepositoryConfigModel remoteRepositoryConfigModel) {
            remoteRepositoryConfigModel.getAdvanced().getNetwork().setDisableProxy(disableProxy);
        } else if (model instanceof FederatedRepositoryConfigModel federatedRepositoryConfigModel) {
            federatedRepositoryConfigModel.getAdvanced().setDisableProxy(disableProxy);
        } else {
            Assert.fail(String.format("failed to set disableProxy '%b' to model '%s'", disableProxy,
                    model.getClass().getSimpleName()));
        }
    }

    @Test
    public void testBuildRemoteDescriptor() {
        typeSpecificConfigModel = new NpmTypeSpecificConfigModel();
        final RemoteRepositoryConfigModel model = prepareRemoteModel();
        RemoteRepoConfig remoteRepoConfig = repoConfigBuilder.buildRemoteRepoConfig(model);
        assertNotNull(remoteRepoConfig.getRepoTypeConfig().getCdnRedirectRepoConfig());
        assertTrue(remoteRepoConfig.getRepoTypeConfig().getCdnRedirectRepoConfig().isEnabled());
        long metadataRetrievalTimeoutSecs = model.getAdvanced().getCache().getMetadataRetrievalTimeoutSecs();
        assertEquals(remoteRepoConfig.getRepoTypeConfig().getMetadataRetrievalTimeoutSecs(),
                metadataRetrievalTimeoutSecs);
    }

    @Test
    public void testBuildDockerRemoteDescriptorWithNewCacheRetrievalPeriod() {
        typeSpecificConfigModel = new DockerTypeSpecificConfigModel();
        final RemoteRepositoryConfigModel model = prepareRemoteModel();
        RemoteRepoConfig remoteRepoConfig = repoConfigBuilder.buildRemoteRepoConfig(model);
        assertEquals(remoteRepoConfig.getRepoTypeConfig().getRetrievalCachePeriodSecs(),
                DEFAULT_DOCKER_REMOTE_RETRIEVAL_CACHE_PERIOD);
    }

    @Test
    public void testPopulateVirtualBasicDescriptorValues() {
        VirtualBasicRepositoryConfigModel model = mock(VirtualBasicRepositoryConfigModel.class);
        String defaultDeploymentRepoName = "defaultDeploymentRepoName";
        when(model.getDefaultDeploymentRepo()).thenReturn(defaultDeploymentRepoName);
        when(centralConfigService.getDescriptor()).thenReturn(mock(CentralConfigDescriptor.class));
        VirtualRepoConfig virtualRepoConfig = VirtualRepoConfig.create("repoKey", RepoType.Generic);
        FederatedRepoConfig federatedRepoConfig = createFederatedRepoConfig(defaultDeploymentRepoName);
        when(repositoryConfigService.findRepositoriesByType(FEDERATED)).thenReturn(List.of(federatedRepoConfig));
        when(repositoryConfigService.findRepositoriesByType(LOCAL)).thenReturn(List.of());
        repoConfigBuilder.populateVirtualBasicConfigValues(model, virtualRepoConfig);
        assertEquals(virtualRepoConfig.getRepoTypeConfig().getDefaultDeploymentRepoRef(), defaultDeploymentRepoName);
    }

    @NotNull
    private FederatedRepoConfig createFederatedRepoConfig(String repoName) {
        FederatedRepoConfig federatedRepoConfig = mock(FederatedRepoConfig.class, RETURNS_DEEP_STUBS);
        when(federatedRepoConfig.getKey()).thenReturn(repoName);
        when(federatedRepoConfig.getPackageType()).thenReturn(RepoType.Generic);
        when(federatedRepoConfig.getRepoTypeConfig().getSnapshotVersionBehavior()).thenReturn(
                SnapshotVersionBehavior.DEPLOYER);
        when(federatedRepoConfig.getRepoTypeConfig().getChecksumPolicyType()).thenReturn(
                LocalRepoChecksumPolicyType.CLIENT);
        return federatedRepoConfig;
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testDockerWithoutStoreArtifactsLocally() {
        typeSpecificConfigModel = new DockerTypeSpecificConfigModel();
        final RemoteRepositoryConfigModel model = prepareRemoteModel();
        model.getAdvanced().setStoreArtifactsLocally(false);
        when(repoConfigBuilder.buildRemoteRepoConfig(model)).thenThrow(new IllegalArgumentException());
    }

    private void fillBasic(BasicRepositoryConfigModel basic) {
        basic.setPublicDescription("");
        basic.setInternalDescription("");
        basic.setIncludesPattern("");
        basic.setExcludesPattern("");
    }

    private LocalRepositoryConfigModel prepareLocalModel() {
        final LocalRepositoryConfigModel model = new LocalRepositoryConfigModel();
        CdnRedirectRepoConfigModel cdnRedirectRepoConfig = new CdnRedirectRepoConfigModel();
        cdnRedirectRepoConfig.setEnabled(true);
        final LocalAdvancedRepositoryConfigModel advanced = new LocalAdvancedRepositoryConfigModel();
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final LocalBasicRepositoryConfigModel basic = new LocalBasicRepositoryConfigModel();
        fillBasic(basic);
        general.setRepoKey("repo");
        advanced.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);

        return model;
    }

    private FederatedRepositoryConfigModel prepareFederatedRepoModel() {
        final FederatedRepositoryConfigModel model = new FederatedRepositoryConfigModel();
        final FederatedRepoAdvancedRepositoryConfigModel advanced = new FederatedRepoAdvancedRepositoryConfigModel();
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final LocalBasicRepositoryConfigModel basic = new LocalBasicRepositoryConfigModel();
        fillBasic(basic);
        general.setRepoKey("repo");
        model.setFederation(new FederationConfigModel());
        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);
        return model;
    }

    private RemoteRepositoryConfigModel prepareRemoteModel() {
        final RemoteRepositoryConfigModel model = new RemoteRepositoryConfigModel();
        CdnRedirectRepoConfigModel cdnRedirectRepoConfig = new CdnRedirectRepoConfigModel();
        cdnRedirectRepoConfig.setEnabled(true);
        final RemoteAdvancedRepositoryConfigModel advanced = new RemoteAdvancedRepositoryConfigModel();
        advanced.setCache(new RemoteCacheRepositoryConfigModel());
        advanced.setNetwork(new RemoteNetworkRepositoryConfigModel());
        final GeneralRepositoryConfigModel general = new GeneralRepositoryConfigModel();
        final RemoteBasicRepositoryConfigModel basic = new RemoteBasicRepositoryConfigModel();
        fillBasic(basic);
        general.setRepoKey("repo");
        advanced.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
        model.setTypeSpecific(typeSpecificConfigModel);
        model.setAdvanced(advanced);
        model.setGeneral(general);
        model.setBasic(basic);
        when(artifactorySystemProperties.getProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    if (constantValue == ConstantValues.debianDefaultArchitectures) {
                        return "deb";
                    }
                    return constantValue.getDefValue().trim();
                });

        return model;
    }
}