package org.artifactory.rest.resource.repositories;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.KeyPairDescriptor;
import org.artifactory.model.RepoConfig;
import org.jfrog.security.common.AuthorizationRoles;

import javax.annotation.security.RolesAllowed;
import javax.ws.rs.Consumes;
import javax.ws.rs.DELETE;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.Objects;

import static javax.ws.rs.core.Response.Status.BAD_REQUEST;

/**
 * A resource to manage all repository key pairs
 *
 * @author Omri Ziv
 */
@RolesAllowed({AuthorizationRoles.ROLE_ADMIN, AuthorizationRoles.ROLE_USER})
public class RepositoryKeyPairsResource {

    private static final String KEY_PAIR_DOES_NOT_EXISTS_MSG = "Key Pair does not exist";
    private final MutableCentralConfigDescriptor configDescriptor;
    private final RepositoryConfigService repositoryConfigService;
    private final String repoKey;

    public RepositoryKeyPairsResource(CentralConfigService centralConfigService,
            RepositoryConfigService repositoryConfigService, String repoKey) {
        this.configDescriptor = centralConfigService.getMutableDescriptor();
        this.repositoryConfigService = repositoryConfigService;
        this.repoKey = repoKey;
    }

    @POST
    @Produces(MediaType.APPLICATION_JSON)
    @Path("promote")
    public Response promote() {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        if (Objects.isNull(repoConfig.getSecurityConfig().getSecondaryKeyPairRef())) {
            return Response.status(BAD_REQUEST)
                    .entity("Nothing to promote")
                    .build();
        }
        repoConfig.getSecurityConfig().promoteKeyPair();
        repositoryConfigService.upsertRepository(repoConfig);
        return Response.ok()
                .entity("Secondary Key Pair of repo '" + repoKey + "' promoted.")
                .build();
    }

    @POST
    @Consumes(MediaType.TEXT_PLAIN)
    @Produces(MediaType.APPLICATION_JSON)
    public Response setKey(String keyName) {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        KeyPairDescriptor keyPairDescriptor = configDescriptor.getKeyPair(keyName);
        if (Objects.isNull(keyPairDescriptor)) {
            return Response.status(BAD_REQUEST)
                    .entity(KEY_PAIR_DOES_NOT_EXISTS_MSG)
                    .build();
        }
        if (keyPairDescriptor.getKey().equals(repoConfig.getSecurityConfig().getPrimaryKeyPairRef())) {
            return Response.ok()
                    .entity(String.format("Key Pair '%s' of repo '%s' was set as primary.", keyName, repoKey))
                    .build();
        }
        if (Objects.isNull(repoConfig.getSecurityConfig().getPrimaryKeyPairRef())) {
            repoConfig.getSecurityConfig().setPrimaryKeyPairRef(keyPairDescriptor.getKey());
            repositoryConfigService.upsertRepository(repoConfig);
            return Response.ok()
                    .entity(String.format("Key Pair '%s' of repo '%s' was set as primary.", keyName, repoKey))
                    .build();
        } else {
            repoConfig.getSecurityConfig().setSecondaryKeyPairRef(keyPairDescriptor.getKey());
            repositoryConfigService.upsertRepository(repoConfig);
            return Response.ok()
                    .entity(String.format("Key Pair '%s' of repo '%s' was set as secondary.", keyName, repoKey))
                    .build();
        }
    }

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    @Path("{keyName: .+}")
    public Response deleteKey(@PathParam("keyName") String keyName) {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        KeyPairDescriptor keyPairDescriptor = configDescriptor.getKeyPair(keyName);
        if (Objects.isNull(keyPairDescriptor)) {
            return Response.status(BAD_REQUEST)
                    .entity(KEY_PAIR_DOES_NOT_EXISTS_MSG)
                    .build();
        }
        if (keyPairDescriptor.getKey().equals(repoConfig.getSecurityConfig().getPrimaryKeyPairRef())) {
            repoConfig.getSecurityConfig().setPrimaryKeyPairRef(null);
            repoConfig.getSecurityConfig().promoteKeyPair();
            repositoryConfigService.upsertRepository(repoConfig);
            return Response.ok()
                    .entity(String.format("Primary Key Pair '%s' of repo '%s' removed.", keyName, repoKey))
                    .build();

        } else if (keyPairDescriptor.getKey().equals(repoConfig.getSecurityConfig().getSecondaryKeyPairRef())) {
            repoConfig.getSecurityConfig().setSecondaryKeyPairRef(null);
            repositoryConfigService.upsertRepository(repoConfig);
            return Response.ok()
                    .entity(String.format("Secondary Key Pair '%s' of repo '%s' removed.", keyName, repoKey))
                    .build();
        } else {
            return Response.status(BAD_REQUEST)
                    .entity("Repo '" + repoKey + "' Not containing Key Pair '" + keyName + "'")
                    .build();
        }
    }

    @POST
    @Consumes(MediaType.TEXT_PLAIN)
    @Produces(MediaType.APPLICATION_JSON)
    @Path("primary")
    public Response setPrimary(String primaryKeyName) {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        KeyPairDescriptor keyPairDescriptor = configDescriptor.getKeyPair(primaryKeyName);
        if (Objects.isNull(keyPairDescriptor)) {
            return Response.status(BAD_REQUEST)
                    .entity(KEY_PAIR_DOES_NOT_EXISTS_MSG)
                    .build();
        }
        repoConfig.getSecurityConfig().setPrimaryKeyPairRef(keyPairDescriptor.getKey());
        repositoryConfigService.upsertRepository(repoConfig);
        return Response.ok()
                .entity("Key Pair '" + primaryKeyName + "' of repo '" + repoKey + "' was set as primary.")
                .build();
    }

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    @Path("primary")
    public Response deletePrimary() {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        repoConfig.getSecurityConfig().setPrimaryKeyPairRef(null);
        repoConfig.getSecurityConfig().promoteKeyPair();
        repositoryConfigService.upsertRepository(repoConfig);
        return Response.ok()
                .entity("Primary Key Pair of repo '" + repoKey + "' deleted successfully.")
                .build();
    }


    @POST
    @Consumes(MediaType.TEXT_PLAIN)
    @Produces(MediaType.APPLICATION_JSON)
    @Path("secondary")
    public Response setSecondary(String secondaryKeyName) {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        KeyPairDescriptor keyPairDescriptor = configDescriptor.getKeyPair(secondaryKeyName);
        if (Objects.isNull(keyPairDescriptor)) {
            return Response.status(BAD_REQUEST)
                    .entity(KEY_PAIR_DOES_NOT_EXISTS_MSG)
                    .build();
        }
        if (Objects.isNull(repoConfig.getSecurityConfig().getPrimaryKeyPairRef())) {
            return Response.status(BAD_REQUEST)
                    .entity("Could not set secondary while primary is not set")
                    .build();
        }
        repoConfig.getSecurityConfig().setSecondaryKeyPairRef(keyPairDescriptor.getKey());
        repositoryConfigService.upsertRepository(repoConfig);
        return Response.ok()
                .entity("Key Pair '" + secondaryKeyName + "' of repo '" + repoKey + "' was set as secondary.")
                .build();
    }

    @DELETE
    @Produces(MediaType.APPLICATION_JSON)
    @Path("secondary")
    public Response deleteSecondary() {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        repoConfig.getSecurityConfig().setSecondaryKeyPairRef(null);
        repositoryConfigService.upsertRepository(repoConfig);
        return Response.ok()
                .entity("Secondary Key Pair of repo '" + repoKey + "' deleted successfully.")
                .build();
    }

    @GET
    @Path("primary/public")
    @Produces(MediaType.TEXT_PLAIN)
    @RolesAllowed({AuthorizationService.ROLE_USER, AuthorizationService.ROLE_ADMIN})
    public Response getPrimaryPublicKey() {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        String primaryKeyPairRef = repoConfig.getSecurityConfig().getPrimaryKeyPairRef();
        if (Objects.isNull(primaryKeyPairRef)) {
            return Response.status(BAD_REQUEST)
                    .entity("Repo '" + repoKey + "' Not containing primary Key Pair")
                    .build();

        }
        KeyPairDescriptor keyPair = configDescriptor.getKeyPair(primaryKeyPairRef);
        if (Objects.isNull(keyPair)) {
            return Response.status(BAD_REQUEST).entity(KEY_PAIR_DOES_NOT_EXISTS_MSG).build();
        }

        return Response.ok()
                .header("Content-Disposition",
                        "attachment; filename=" + keyPair.getAliasWithExtension())
                .entity(keyPair.getPublicKey()).build();
    }

    @GET
    @Path("secondary/public")
    @Produces(MediaType.TEXT_PLAIN)
    @RolesAllowed({AuthorizationService.ROLE_USER, AuthorizationService.ROLE_ADMIN})
    public Response getSecondaryPublicKey() {
        RepoConfig repoConfig = getRepoConfig(repoKey);
        if (Objects.isNull(repoConfig)) {
            return Response.status(BAD_REQUEST).build();
        }
        String secondaryKeyPairRef = repoConfig.getSecurityConfig().getSecondaryKeyPairRef();
        if (Objects.isNull(secondaryKeyPairRef)) {
            return Response.status(BAD_REQUEST)
                    .entity("Repo '" + repoKey + "' Not containing secondary Key Pair")
                    .build();

        }
        KeyPairDescriptor keyPair = configDescriptor.getKeyPair(secondaryKeyPairRef);
        if (Objects.isNull(keyPair)) {
            return Response.status(BAD_REQUEST).entity(KEY_PAIR_DOES_NOT_EXISTS_MSG).build();
        }
        return Response.ok()
                .header("Content-Disposition",
                        "attachment; filename=" + keyPair.getAliasWithExtension())
                .entity(keyPair.getPublicKey()).build();
    }

    private RepoConfig getRepoConfig(String repoKey) {
        try {
            return repositoryConfigService.getRepoConfigByKey(repoKey);
        } catch (Exception e) {
            return null;
        }
    }
}
