package org.artifactory.rest.resource.servicetrust;

import org.artifactory.servicetrust.ServiceTrustService;
import org.artifactory.servicetrust.ServiceTrustServiceImpl;
import org.artifactory.servicetrust.ServiceTrustUrls;
import org.artifactory.servicetrust.ServiceTrustUseCase;
import org.artifactory.servicetrust.ServiceTrustUseCaseData;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.common.ServiceId;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import javax.ws.rs.core.Response;
import java.util.Optional;

import static org.artifactory.rest.resource.servicetrust.ServiceTrustResourceRegisterTokenTest.assertResponse;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doCallRealMethod;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.AssertJUnit.assertTrue;

public class ServiceTrustResourceCreateTokenTest {


    @Mock
    AccessServiceTestImpl accessService;

    @Mock
    private AccessClient accessClient;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.openMocks(this);
    }

    @Test
    public void whenUseCaseIsUnknownThen404() {
        ServiceTrustResource resource = new ServiceTrustResourceForTest();
        ServiceTrustService service = Mockito.mock(ServiceTrustService.class);
        when(service.findUseCase(any())).thenReturn(Optional.empty());
        resource.setServiceTrustService(service);

        Response response = resource.createToken("unknown");

        assertResponse(response, 404, "Unknown use case");
    }

    @Test
    public void whenUseCaseIsOkThen200() {
        ServiceTrustResource resource = new ServiceTrustResourceForTest();
        ServiceTrustService service = Mockito.mock(ServiceTrustService.class);
        when(service.findUseCase(any())).thenReturn(Optional.of(ServiceTrustUseCase.ARTIFACTORY_COLD));
        when(service.getPairingToken(any(), any())).thenReturn("token");
        resource.setServiceTrustService(service);

        Response response = resource.createToken("artifactory-cold");

        int status = response.getStatus();
        String token = ((PairingToken) response.getEntity()).getPairingToken();
        assertEquals(status, 200);
        assertEquals(token, "token");
    }

    @Test
    public void whenUseCaseFedRepoGivenArtifactoryPrefixInHostnameThen200() {

        ServiceTrustResource realResource = new ServiceTrustResourceForTest();
        ServiceTrustResource resource = spy(realResource);
        ServiceTrustServiceImpl service = spy(ServiceTrustServiceImpl.class);

        when(accessService.getArtifactoryServiceId()).thenReturn(new ServiceId("jfrt", "test"));
        when(accessService.getAccessClient()).thenReturn(accessClient);
        when(accessService.createToken(any())).thenCallRealMethod();
        service.setAccessService(accessService);
        doReturn("https://artifactory.mycompany.com/artifactory").when(resource).getFederatedRepoUrlBase();

        doReturn(Optional.of(ServiceTrustUseCase.FEDERATED_REPO)).when(service).findUseCase(anyString());
        doCallRealMethod().when(service)
                .getPairingToken(any(ServiceTrustUseCaseData.class), any(ServiceTrustUrls.class));
        resource.setServiceTrustService(service);

        Response response = resource.createTokenResponse("federated-repo/test", ServiceTrustUseCase.FEDERATED_REPO);

        int status = response.getStatus();
        String token = ((PairingToken) response.getEntity()).getPairingToken();
        assertEquals(status, 200);
        assertTrue(token.contains("https://artifactory.mycompany.com"));
        assertFalse(token.contains("https://artifactory.mycompany.com/artifactory"));

        doReturn("https://artifactory.mycompany.com").when(resource).getFederatedRepoUrlBase();

        Response response2 = resource.createTokenResponse("federated-repo/test2", ServiceTrustUseCase.FEDERATED_REPO);

        String token2 = ((PairingToken) response2.getEntity()).getPairingToken();
        assertTrue(token2.contains("https://artifactory.mycompany.com"));
    }


    private static class ServiceTrustResourceForTest extends ServiceTrustResource {
        @Override
        String getRestApiUrl() {
            return "";
        }

        @Override
        String getBaseUrl() {
            return "";
        }
    }
}